from __future__ import annotations

from typing import Any

import pygame

from ..entities_constants import (
    CAR_HEIGHT,
    CAR_WIDTH,
    FLASHLIGHT_HEIGHT,
    FLASHLIGHT_WIDTH,
    FUEL_CAN_HEIGHT,
    FUEL_CAN_WIDTH,
    HUMANOID_RADIUS,
    SHOES_HEIGHT,
    SHOES_WIDTH,
    SURVIVOR_APPROACH_RADIUS,
    SURVIVOR_MAX_SAFE_PASSENGERS,
)
from .constants import (
    CAR_ZOMBIE_DAMAGE,
    FUEL_HINT_DURATION_MS,
    SURVIVOR_OVERLOAD_DAMAGE_RATIO,
)
from ..localization import translate as tr
from ..models import GameData
from ..rng import get_rng
from ..entities import Car
from .footprints import get_shrunk_sprite
from .spawn import maintain_waiting_car_supply
from .survivors import (
    add_survivor_message,
    apply_passenger_speed_penalty,
    drop_survivors_from_car,
    handle_survivor_zombie_collisions,
    increase_survivor_capacity,
    respawn_buddies_near_player,
)
from .utils import rect_visible_on_screen
from .ambient import sync_ambient_palette_with_flashlights


def _interaction_radius(width: float, height: float) -> float:
    """Approximate interaction reach for a humanoid and an object."""
    return HUMANOID_RADIUS + (width + height) / 4


RNG = get_rng()


def check_interactions(game_data: GameData, config: dict[str, Any]) -> None:
    """Check and handle interactions between entities."""
    player = game_data.player
    assert player is not None
    car = game_data.car
    zombie_group = game_data.groups.zombie_group
    all_sprites = game_data.groups.all_sprites
    survivor_group = game_data.groups.survivor_group
    state = game_data.state
    walkable_cells = game_data.layout.walkable_cells
    outside_cells = game_data.layout.outside_cells
    fuel = game_data.fuel
    flashlights = game_data.flashlights or []
    shoes_list = game_data.shoes or []
    camera = game_data.camera
    stage = game_data.stage
    cell_size = game_data.cell_size
    maintain_waiting_car_supply(game_data)
    active_car = car if car and car.alive() else None
    waiting_cars = game_data.waiting_cars
    shrunk_car = get_shrunk_sprite(active_car, 0.8) if active_car else None

    car_interaction_radius = _interaction_radius(CAR_WIDTH, CAR_HEIGHT)
    fuel_interaction_radius = _interaction_radius(FUEL_CAN_WIDTH, FUEL_CAN_HEIGHT)
    flashlight_interaction_radius = _interaction_radius(FLASHLIGHT_WIDTH, FLASHLIGHT_HEIGHT)
    shoes_interaction_radius = _interaction_radius(SHOES_WIDTH, SHOES_HEIGHT)

    def _rect_center_cell(rect: pygame.Rect) -> tuple[int, int] | None:
        if cell_size <= 0:
            return None
        return (int(rect.centerx // cell_size), int(rect.centery // cell_size))

    def _cell_center(cell: tuple[int, int]) -> tuple[int, int]:
        return (
            int((cell[0] * cell_size) + (cell_size / 2)),
            int((cell[1] * cell_size) + (cell_size / 2)),
        )

    def _player_near_point(point: tuple[float, float], radius: float) -> bool:
        dx = point[0] - player.x
        dy = point[1] - player.y
        return dx * dx + dy * dy <= radius * radius

    def _player_near_sprite(sprite_obj: pygame.sprite.Sprite | None, radius: float) -> bool:
        return bool(sprite_obj and sprite_obj.alive() and _player_near_point(sprite_obj.rect.center, radius))

    def _player_near_car(car_obj: Car | None) -> bool:
        return _player_near_sprite(car_obj, car_interaction_radius)

    # Fuel pickup
    if fuel and fuel.alive() and not state.has_fuel and not player.in_car:
        if _player_near_point(fuel.rect.center, fuel_interaction_radius):
            state.has_fuel = True
            state.fuel_message_until = 0
            state.hint_expires_at = 0
            state.hint_target_type = None
            fuel.kill()
            game_data.fuel = None
            print("Fuel acquired!")

    # Flashlight pickup
    if not player.in_car:
        for flashlight in list(flashlights):
            if not flashlight.alive():
                continue
            if _player_near_point(flashlight.rect.center, flashlight_interaction_radius):
                state.flashlight_count += 1
                state.hint_expires_at = 0
                state.hint_target_type = None
                flashlight.kill()
                try:
                    flashlights.remove(flashlight)
                except ValueError:
                    pass
                print("Flashlight acquired!")
                break

        for shoes in list(shoes_list):
            if not shoes.alive():
                continue
            if _player_near_point(shoes.rect.center, shoes_interaction_radius):
                state.shoes_count += 1
                state.hint_expires_at = 0
                state.hint_target_type = None
                shoes.kill()
                try:
                    shoes_list.remove(shoes)
                except ValueError:
                    pass
                print("Shoes acquired!")
                break

    sync_ambient_palette_with_flashlights(game_data)

    buddies = [
        survivor for survivor in survivor_group if survivor.alive() and survivor.is_buddy and not survivor.rescued
    ]

    # Buddy interactions (Stage 3)
    if stage.buddy_required_count > 0 and buddies:
        for buddy in list(buddies):
            if not buddy.alive():
                continue
            buddy_on_screen = rect_visible_on_screen(camera, buddy.rect)
            if not player.in_car:
                dist_to_player_sq = (player.x - buddy.x) ** 2 + (player.y - buddy.y) ** 2
                if dist_to_player_sq <= SURVIVOR_APPROACH_RADIUS * SURVIVOR_APPROACH_RADIUS:
                    buddy.set_following()
            elif player.in_car and active_car and shrunk_car:
                g = pygame.sprite.Group()
                g.add(buddy)
                if pygame.sprite.spritecollide(shrunk_car, g, False, pygame.sprite.collide_circle):
                    prospective_passengers = state.survivors_onboard + 1
                    capacity_limit = state.survivor_capacity
                    if prospective_passengers > capacity_limit:
                        overload_damage = max(
                            1,
                            int(active_car.max_health * SURVIVOR_OVERLOAD_DAMAGE_RATIO),
                        )
                        add_survivor_message(game_data, tr("survivors.too_many_aboard"))
                        active_car._take_damage(overload_damage)
                    state.buddy_onboard += 1
                    buddy.kill()
                    continue

            if buddy.alive() and pygame.sprite.spritecollide(buddy, zombie_group, False, pygame.sprite.collide_circle):
                if buddy_on_screen:
                    state.game_over_message = tr("game_over.scream")
                    state.game_over = True
                    state.game_over_at = state.game_over_at or pygame.time.get_ticks()
                else:
                    if walkable_cells:
                        new_cell = RNG.choice(walkable_cells)
                        buddy.teleport(_cell_center(new_cell))
                    else:
                        buddy.teleport((game_data.level_width // 2, game_data.level_height // 2))
                    buddy.following = False

    # Player entering an active car already under control
    if not player.in_car and _player_near_car(active_car) and active_car and active_car.health > 0:
        if state.has_fuel:
            player.in_car = True
            all_sprites.remove(player)
            state.hint_expires_at = 0
            state.hint_target_type = None
            print("Player entered car!")
        else:
            if not stage.endurance_stage:
                now_ms = state.elapsed_play_ms
                state.fuel_message_until = now_ms + FUEL_HINT_DURATION_MS
                state.hint_target_type = "fuel"

    # Claim a waiting/parked car when the player finally reaches it
    if not player.in_car and not active_car and waiting_cars:
        claimed_car: Car | None = None
        for parked_car in waiting_cars:
            if _player_near_car(parked_car):
                claimed_car = parked_car
                break
        if claimed_car:
            if state.has_fuel:
                try:
                    game_data.waiting_cars.remove(claimed_car)
                except ValueError:
                    pass
                game_data.car = claimed_car
                active_car = claimed_car
                player.in_car = True
                all_sprites.remove(player)
                state.hint_expires_at = 0
                state.hint_target_type = None
                apply_passenger_speed_penalty(game_data)
                maintain_waiting_car_supply(game_data)
                print("Player claimed a waiting car!")
            else:
                if not stage.endurance_stage:
                    now_ms = state.elapsed_play_ms
                    state.fuel_message_until = now_ms + FUEL_HINT_DURATION_MS
                    state.hint_target_type = "fuel"

    # Bonus: collide a parked car while driving to repair/extend capabilities
    if player.in_car and active_car and shrunk_car and waiting_cars:
        waiting_group = pygame.sprite.Group(waiting_cars)
        collided_waiters = pygame.sprite.spritecollide(shrunk_car, waiting_group, False, pygame.sprite.collide_rect)
        if collided_waiters:
            removed_any = False
            capacity_increments = 0
            for parked in collided_waiters:
                if not parked.alive():
                    continue
                parked.kill()
                try:
                    game_data.waiting_cars.remove(parked)
                except ValueError:
                    pass
                active_car.health = active_car.max_health
                active_car._update_color()
                removed_any = True
                if stage.rescue_stage:
                    capacity_increments += 1
            if removed_any:
                if capacity_increments:
                    increase_survivor_capacity(game_data, capacity_increments)
                maintain_waiting_car_supply(game_data)

    # Car hitting zombies
    if player.in_car and active_car and active_car.health > 0 and shrunk_car:
        zombies_hit = pygame.sprite.spritecollide(shrunk_car, zombie_group, True)
        if zombies_hit:
            active_car._take_damage(CAR_ZOMBIE_DAMAGE * len(zombies_hit))

    if stage.rescue_stage and player.in_car and active_car and shrunk_car and survivor_group:
        boarded = pygame.sprite.spritecollide(shrunk_car, survivor_group, True, pygame.sprite.collide_circle)
        if boarded:
            state.survivors_onboard += len(boarded)
            apply_passenger_speed_penalty(game_data)
            capacity_limit = state.survivor_capacity
            if state.survivors_onboard > capacity_limit:
                overload_damage = max(
                    1,
                    int(active_car.max_health * SURVIVOR_OVERLOAD_DAMAGE_RATIO),
                )
                add_survivor_message(game_data, tr("survivors.too_many_aboard"))
                active_car._take_damage(overload_damage)

    if stage.rescue_stage:
        handle_survivor_zombie_collisions(game_data, config)

    # Handle car destruction
    if car and car.alive() and car.health <= 0:
        car_destroyed_pos = car.rect.center
        car.kill()
        if stage.rescue_stage:
            drop_survivors_from_car(game_data, car_destroyed_pos)
        if player.in_car:
            player.in_car = False
            player.x, player.y = car_destroyed_pos[0], car_destroyed_pos[1]
            player.rect.center = (int(player.x), int(player.y))
            if player not in all_sprites:
                all_sprites.add(player, layer=2)
            print("Car destroyed! Player ejected.")

        # Clear active car and let the player hunt for another waiting car.
        game_data.car = None
        state.survivor_capacity = SURVIVOR_MAX_SAFE_PASSENGERS
        apply_passenger_speed_penalty(game_data)

        # Bring back the buddies near the player after losing the car
        respawn_buddies_near_player(game_data)
        maintain_waiting_car_supply(game_data)

    # Player getting caught by zombies
    if not player.in_car and player in all_sprites:
        shrunk_player = get_shrunk_sprite(player, 0.8)
        collisions = pygame.sprite.spritecollide(shrunk_player, zombie_group, False, pygame.sprite.collide_circle)
        if any(not zombie.carbonized for zombie in collisions):
            if not state.game_over:
                state.game_over = True
                state.game_over_at = pygame.time.get_ticks()
                state.game_over_message = tr("game_over.scream")

    # Player escaping on foot after dawn (Stage 5)
    if (
        stage.endurance_stage
        and state.dawn_ready
        and not player.in_car
        and outside_cells
        and (player_cell := _rect_center_cell(player.rect)) is not None
        and player_cell in outside_cells
    ):
        state.game_won = True

    # Player escaping the level
    if player.in_car and car and car.alive() and state.has_fuel:
        buddy_ready = True
        if stage.buddy_required_count > 0:
            buddy_ready = state.buddy_onboard >= stage.buddy_required_count
        car_cell = _rect_center_cell(car.rect)
        if buddy_ready and car_cell is not None and car_cell in outside_cells:
            if stage.buddy_required_count > 0:
                state.buddy_rescued = min(stage.buddy_required_count, state.buddy_onboard)
            if stage.rescue_stage and state.survivors_onboard:
                state.survivors_rescued += state.survivors_onboard
                state.survivors_onboard = 0
                apply_passenger_speed_penalty(game_data)
            state.game_won = True

    return None



from __future__ import annotations

import math
from typing import Any, Sequence

import pygame

from ..entities import (
    Car,
    Player,
    Survivor,
    Wall,
    Zombie,
)
from ..entities_constants import (
    HUMANOID_WALL_BUMP_FRAMES,
    PLAYER_SPEED,
    ZOMBIE_SEPARATION_DISTANCE,
    ZOMBIE_WALL_HUG_SENSOR_DISTANCE,
)
from ..gameplay_constants import (
    SHOES_SPEED_MULTIPLIER_ONE,
    SHOES_SPEED_MULTIPLIER_TWO,
)
from ..models import FallingZombie, GameData
from ..world_grid import WallIndex, apply_tile_edge_nudge, walls_for_radius
from .constants import MAX_ZOMBIES
from .spawn import spawn_weighted_zombie, update_falling_zombies
from .survivors import update_survivors
from .utils import rect_visible_on_screen


def process_player_input(
    keys: Sequence[bool],
    player: Player,
    car: Car | None,
    shoes_count: int = 0,
    pad_input: tuple[float, float] = (0.0, 0.0),
) -> tuple[float, float, float, float]:
    """Process keyboard input and return movement deltas."""
    dx_input, dy_input = 0, 0
    if keys[pygame.K_w] or keys[pygame.K_UP]:
        dy_input -= 1
    if keys[pygame.K_s] or keys[pygame.K_DOWN]:
        dy_input += 1
    if keys[pygame.K_a] or keys[pygame.K_LEFT]:
        dx_input -= 1
    if keys[pygame.K_d] or keys[pygame.K_RIGHT]:
        dx_input += 1
    dx_input += pad_input[0]
    dy_input += pad_input[1]

    player.update_facing_from_input(dx_input, dy_input)

    player_dx, player_dy, car_dx, car_dy = 0, 0, 0, 0

    if player.in_car and car and car.alive():
        car.update_facing_from_input(dx_input, dy_input)
        target_speed = car.speed
        move_len = math.hypot(dx_input, dy_input)
        if move_len > 0:
            car_dx, car_dy = (
                (dx_input / move_len) * target_speed,
                (dy_input / move_len) * target_speed,
            )
    elif not player.in_car:
        target_speed = PLAYER_SPEED * _shoes_speed_multiplier(shoes_count)
        move_len = math.hypot(dx_input, dy_input)
        if move_len > 0:
            player_dx, player_dy = (
                (dx_input / move_len) * target_speed,
                (dy_input / move_len) * target_speed,
            )

    return player_dx, player_dy, car_dx, car_dy


def _shoes_speed_multiplier(shoes_count: int) -> float:
    count = max(0, int(shoes_count))
    if count >= 2:
        return SHOES_SPEED_MULTIPLIER_TWO
    if count == 1:
        return SHOES_SPEED_MULTIPLIER_ONE
    return 1.0


def _handle_pitfall_detection(
    x: float,
    y: float,
    cell_size: int,
    pitfall_cells: set[tuple[int, int]],
    pull_distance: float,
) -> tuple[int, int] | None:
    """Check if position is in pitfall and return pulled target coordinates if so."""
    cx, cy = int(x // cell_size), int(y // cell_size)
    if (cx, cy) not in pitfall_cells:
        return None

    cell_center_x = (cx * cell_size) + (cell_size // 2)
    cell_center_y = (cy * cell_size) + (cell_size // 2)
    dx, dy = cell_center_x - x, cell_center_y - y
    dist = math.hypot(dx, dy)
    if dist > 0:
        move_factor = min(1.0, pull_distance / dist)
        return int(x + dx * move_factor), int(y + dy * move_factor)
    return int(x), int(y)


def update_entities(
    game_data: GameData,
    player_dx: float,
    player_dy: float,
    car_dx: float,
    car_dy: float,
    config: dict[str, Any],
    wall_index: WallIndex | None = None,
) -> None:
    """Update positions and states of game entities."""
    player = game_data.player
    assert player is not None
    car = game_data.car
    wall_group = game_data.groups.wall_group
    all_sprites = game_data.groups.all_sprites
    zombie_group = game_data.groups.zombie_group
    survivor_group = game_data.groups.survivor_group
    camera = game_data.camera
    stage = game_data.stage
    active_car = car if car and car.alive() else None
    wall_cells = game_data.layout.wall_cells
    pitfall_cells = game_data.layout.pitfall_cells
    walkable_cells = game_data.layout.walkable_cells
    bevel_corners = game_data.layout.bevel_corners

    all_walls = list(wall_group) if wall_index is None else None

    def _walls_near(center: tuple[float, float], radius: float) -> list[Wall]:
        if wall_index is None:
            return all_walls or []
        return walls_for_radius(
            wall_index,
            center,
            radius,
            cell_size=game_data.cell_size,
            grid_cols=stage.grid_cols,
            grid_rows=stage.grid_rows,
        )

    # Update player/car movement
    if player.in_car and active_car:
        car_dx, car_dy = apply_tile_edge_nudge(
            active_car.x,
            active_car.y,
            car_dx,
            car_dy,
            cell_size=game_data.cell_size,
            wall_cells=wall_cells,
            bevel_corners=bevel_corners,
            grid_cols=stage.grid_cols,
            grid_rows=stage.grid_rows,
        )
        car_walls = _walls_near((active_car.x, active_car.y), 150.0)
        active_car.move(
            car_dx,
            car_dy,
            car_walls,
            walls_nearby=wall_index is not None,
            cell_size=game_data.cell_size,
            pitfall_cells=pitfall_cells,
        )
        player.rect.center = active_car.rect.center
        player.x, player.y = active_car.x, active_car.y
    elif not player.in_car:
        # Ensure player is in all_sprites if not in car
        if player not in all_sprites:
            all_sprites.add(player, layer=2)
        player_dx, player_dy = apply_tile_edge_nudge(
            player.x,
            player.y,
            player_dx,
            player_dy,
            cell_size=game_data.cell_size,
            wall_cells=wall_cells,
            bevel_corners=bevel_corners,
            grid_cols=stage.grid_cols,
            grid_rows=stage.grid_rows,
        )
        player.move(
            player_dx,
            player_dy,
            wall_group,
            wall_index=wall_index,
            cell_size=game_data.cell_size,
            level_width=game_data.level_width,
            level_height=game_data.level_height,
            pitfall_cells=pitfall_cells,
            walkable_cells=walkable_cells,
        )
    else:
        # Player flagged as in-car but car is gone; drop them back to foot control
        player.in_car = False

    # Update camera
    target_for_camera = active_car if player.in_car and active_car else player
    camera.update(target_for_camera)

    if player.inner_wall_hit and player.inner_wall_cell is not None:
        game_data.state.player_wall_target_cell = player.inner_wall_cell
        game_data.state.player_wall_target_ttl = HUMANOID_WALL_BUMP_FRAMES
    elif game_data.state.player_wall_target_ttl > 0:
        game_data.state.player_wall_target_ttl -= 1
        if game_data.state.player_wall_target_ttl <= 0:
            game_data.state.player_wall_target_cell = None

    wall_target_cell = game_data.state.player_wall_target_cell if game_data.state.player_wall_target_ttl > 0 else None

    update_survivors(
        game_data,
        wall_index=wall_index,
        wall_target_cell=wall_target_cell,
    )
    update_falling_zombies(game_data, config)

    # Spawn new zombies if needed
    current_time = pygame.time.get_ticks()
    spawn_interval = max(1, stage.spawn_interval_ms)
    spawn_blocked = stage.endurance_stage and game_data.state.dawn_ready
    if (
        len(zombie_group) < MAX_ZOMBIES
        and not spawn_blocked
        and current_time - game_data.state.last_zombie_spawn_time > spawn_interval
    ):
        if spawn_weighted_zombie(game_data, config):
            game_data.state.last_zombie_spawn_time = current_time

    # Update zombies
    target_center = active_car.rect.center if player.in_car and active_car else player.rect.center
    buddies = [
        survivor for survivor in survivor_group if survivor.alive() and survivor.is_buddy and not survivor.rescued
    ]
    buddies_on_screen = [buddy for buddy in buddies if rect_visible_on_screen(camera, buddy.rect)]

    survivors_on_screen: list[Survivor] = []
    if stage.rescue_stage:
        for survivor in survivor_group:
            if survivor.alive():
                if rect_visible_on_screen(camera, survivor.rect):
                    survivors_on_screen.append(survivor)

    zombies_sorted: list[Zombie] = sorted(list(zombie_group), key=lambda z: z.x)

    def _nearby_zombies(index: int) -> list[Zombie]:
        center = zombies_sorted[index]
        neighbors: list[Zombie] = []
        search_radius = ZOMBIE_SEPARATION_DISTANCE + PLAYER_SPEED
        for left in range(index - 1, -1, -1):
            other = zombies_sorted[left]
            if center.x - other.x > search_radius:
                break
            if other.alive():
                neighbors.append(other)
        for right in range(index + 1, len(zombies_sorted)):
            other = zombies_sorted[right]
            if other.x - center.x > search_radius:
                break
            if other.alive():
                neighbors.append(other)
        return neighbors

    for idx, zombie in enumerate(zombies_sorted):
        target = target_center
        if buddies_on_screen:
            dist_to_target_sq = (target_center[0] - zombie.x) ** 2 + (target_center[1] - zombie.y) ** 2
            nearest_buddy = min(
                buddies_on_screen,
                key=lambda buddy: (buddy.rect.centerx - zombie.x) ** 2 + (buddy.rect.centery - zombie.y) ** 2,
            )
            dist_to_buddy_sq = (nearest_buddy.rect.centerx - zombie.x) ** 2 + (
                nearest_buddy.rect.centery - zombie.y
            ) ** 2
            if dist_to_buddy_sq < dist_to_target_sq:
                target = nearest_buddy.rect.center

        if stage.rescue_stage:
            zombie_on_screen = rect_visible_on_screen(camera, zombie.rect)
            if zombie_on_screen:
                candidate_positions: list[tuple[int, int]] = []
                for survivor in survivors_on_screen:
                    candidate_positions.append(survivor.rect.center)
                for buddy in buddies_on_screen:
                    candidate_positions.append(buddy.rect.center)
                candidate_positions.append(player.rect.center)
                if candidate_positions:
                    target = min(
                        candidate_positions,
                        key=lambda pos: (pos[0] - zombie.x) ** 2 + (pos[1] - zombie.y) ** 2,
                    )
        nearby_candidates = _nearby_zombies(idx)
        zombie_search_radius = ZOMBIE_WALL_HUG_SENSOR_DISTANCE + zombie.radius + 120
        nearby_walls = _walls_near((zombie.x, zombie.y), zombie_search_radius)
        zombie.update(
            target,
            nearby_walls,
            nearby_candidates,
            footprints=game_data.state.footprints,
            cell_size=game_data.cell_size,
            grid_cols=stage.grid_cols,
            grid_rows=stage.grid_rows,
            level_width=game_data.level_width,
            level_height=game_data.level_height,
            outer_wall_cells=game_data.layout.outer_wall_cells,
            wall_cells=game_data.layout.wall_cells,
            pitfall_cells=game_data.layout.pitfall_cells,
            bevel_corners=game_data.layout.bevel_corners,
        )

        # Check zombie pitfall
        pull_dist = zombie.radius * 0.5
        pitfall_target = _handle_pitfall_detection(
            zombie.x,
            zombie.y,
            game_data.cell_size,
            pitfall_cells,
            pull_dist,
        )
        if pitfall_target is not None:
            zombie.kill()
            fall = FallingZombie(
                start_pos=(int(zombie.x), int(zombie.y)),
                target_pos=pitfall_target,
                started_at_ms=pygame.time.get_ticks(),
                pre_fx_ms=0,
                fall_duration_ms=500,
                dust_duration_ms=0,
                tracker=zombie.tracker,
                wall_hugging=zombie.wall_hugging,
                mode="pitfall",
            )
            game_data.state.falling_zombies.append(fall)



from __future__ import annotations

from bisect import bisect_left
from typing import Any

import math

import pygame

from ..entities_constants import (
    BUDDY_RADIUS,
    CAR_SPEED,
    PLAYER_RADIUS,
    SURVIVOR_MAX_SAFE_PASSENGERS,
    SURVIVOR_MIN_SPEED_FACTOR,
    SURVIVOR_RADIUS,
    ZOMBIE_RADIUS,
)
from .constants import SURVIVOR_MESSAGE_DURATION_MS, SURVIVOR_SPEED_PENALTY_PER_PASSENGER
from ..localization import translate_dict, translate_list
from ..models import GameData, ProgressState
from ..rng import get_rng
from ..entities import Survivor, Zombie, spritecollideany_walls
from ..world_grid import WallIndex
from .spawn import _create_zombie
from .utils import find_nearby_offscreen_spawn_position, rect_visible_on_screen

RNG = get_rng()


def update_survivors(
    game_data: GameData,
    wall_index: WallIndex | None = None,
    wall_target_cell: tuple[int, int] | None = None,
) -> None:
    if not (game_data.stage.rescue_stage or game_data.stage.buddy_required_count > 0):
        return
    survivor_group = game_data.groups.survivor_group
    wall_group = game_data.groups.wall_group
    player = game_data.player
    car = game_data.car
    if not player:
        return
    target_rect = car.rect if player.in_car and car and car.alive() else player.rect
    target_pos = target_rect.center
    survivors = [s for s in survivor_group if s.alive()]

    for survivor in survivors:
        survivor.update_behavior(
            target_pos,
            wall_group,
            wall_index=wall_index,
            cell_size=game_data.cell_size,
            wall_cells=game_data.layout.wall_cells,
            pitfall_cells=game_data.layout.pitfall_cells,
            walkable_cells=game_data.layout.walkable_cells,
            bevel_corners=game_data.layout.bevel_corners,
            grid_cols=game_data.stage.grid_cols,
            grid_rows=game_data.stage.grid_rows,
            level_width=game_data.level_width,
            level_height=game_data.level_height,
            wall_target_cell=wall_target_cell,
        )

    # Gently prevent survivors from overlapping the player or each other
    def _separate_from_point(survivor: Survivor, point: tuple[float, float], min_dist: float) -> None:
        dx = point[0] - survivor.x
        dy = point[1] - survivor.y
        dist = math.hypot(dx, dy)
        if dist == 0:
            angle = RNG.uniform(0, math.tau)
            dx, dy = math.cos(angle), math.sin(angle)
            dist = 1
        if dist < min_dist:
            push = min_dist - dist
            survivor.x -= (dx / dist) * push
            survivor.y -= (dy / dist) * push
            survivor.rect.center = (int(survivor.x), int(survivor.y))

    player_overlap = (SURVIVOR_RADIUS + PLAYER_RADIUS) * 1.05
    survivor_overlap = (SURVIVOR_RADIUS * 2) * 1.05

    player_point = (player.x, player.y)
    for survivor in survivors:
        _separate_from_point(survivor, player_point, player_overlap)

    survivors_with_x = sorted(((survivor.x, survivor) for survivor in survivors), key=lambda item: item[0])
    for i, (base_x, survivor) in enumerate(survivors_with_x):
        for other_base_x, other in survivors_with_x[i + 1 :]:
            if other_base_x - base_x > survivor_overlap:
                break
            dx = other.x - survivor.x
            dy = other.y - survivor.y
            dist = math.hypot(dx, dy)
            if dist == 0:
                angle = RNG.uniform(0, math.tau)
                dx, dy = math.cos(angle), math.sin(angle)
                dist = 1
            if dist < survivor_overlap:
                push = (survivor_overlap - dist) / 2
                offset_x = (dx / dist) * push
                offset_y = (dy / dist) * push
                survivor.x -= offset_x
                survivor.y -= offset_y
                other.x += offset_x
                other.y += offset_y
                survivor.rect.center = (int(survivor.x), int(survivor.y))
                other.rect.center = (int(other.x), int(other.y))


def calculate_car_speed_for_passengers(passengers: int, *, capacity: int = SURVIVOR_MAX_SAFE_PASSENGERS) -> float:
    cap = max(1, capacity)
    load_ratio = max(0.0, passengers / cap)
    penalty = SURVIVOR_SPEED_PENALTY_PER_PASSENGER * load_ratio
    penalty = min(0.95, max(0.0, penalty))
    adjusted = CAR_SPEED * (1 - penalty)
    if passengers <= cap:
        return max(CAR_SPEED * SURVIVOR_MIN_SPEED_FACTOR, adjusted)

    overload = passengers - cap
    overload_factor = 1 / math.sqrt(overload + 1)
    overloaded_speed = CAR_SPEED * overload_factor
    return max(CAR_SPEED * SURVIVOR_MIN_SPEED_FACTOR, overloaded_speed)


def apply_passenger_speed_penalty(game_data: GameData) -> None:
    car = game_data.car
    if not car:
        return
    if not game_data.stage.rescue_stage:
        car.speed = CAR_SPEED
        return
    car.speed = calculate_car_speed_for_passengers(
        game_data.state.survivors_onboard,
        capacity=game_data.state.survivor_capacity,
    )


def increase_survivor_capacity(game_data: GameData, increments: int = 1) -> None:
    if increments <= 0:
        return
    if not game_data.stage.rescue_stage:
        return
    state = game_data.state
    state.survivor_capacity += increments * SURVIVOR_MAX_SAFE_PASSENGERS
    apply_passenger_speed_penalty(game_data)


def add_survivor_message(game_data: GameData, text: str) -> None:
    expires = pygame.time.get_ticks() + SURVIVOR_MESSAGE_DURATION_MS
    game_data.state.survivor_messages.append({"text": text, "expires_at": expires})


def _normalize_legacy_conversion_lines(data: dict[str, Any]) -> list[str]:
    numbered: list[tuple[int, str]] = []
    others: list[tuple[str, str]] = []
    for key, value in data.items():
        if not value:
            continue
        text = str(value)
        if isinstance(key, str) and key.startswith("line"):
            suffix = key[4:]
            if suffix.isdigit():
                numbered.append((int(suffix), text))
                continue
        others.append((str(key), text))
    numbered.sort(key=lambda item: item[0])
    others.sort(key=lambda item: item[0])
    return [text for _, text in numbered] + [text for _, text in others]


def _get_survivor_conversion_messages(stage_id: str) -> list[str]:
    key = f"stages.{stage_id}.survivor_conversion_messages"
    raw = translate_list(key)
    if raw:
        return [str(item) for item in raw if item]
    legacy = translate_dict(f"stages.{stage_id}.conversion_lines")
    if legacy:
        return _normalize_legacy_conversion_lines(legacy)
    return []


def random_survivor_conversion_line(stage_id: str) -> str:
    lines = _get_survivor_conversion_messages(stage_id)
    if not lines:
        return ""
    return RNG.choice(lines)


def cleanup_survivor_messages(state: ProgressState) -> None:
    now = pygame.time.get_ticks()
    state.survivor_messages = [msg for msg in state.survivor_messages if msg.get("expires_at", 0) > now]


def drop_survivors_from_car(game_data: GameData, origin: tuple[int, int]) -> None:
    """Respawn boarded survivors back into the world after a crash."""
    count = game_data.state.survivors_onboard
    if count <= 0:
        return
    wall_group = game_data.groups.wall_group
    survivor_group = game_data.groups.survivor_group
    all_sprites = game_data.groups.all_sprites

    for survivor_idx in range(count):
        placed = False
        for attempt in range(6):
            angle = RNG.uniform(0, math.tau)
            dist = RNG.uniform(16, 40)
            pos = (
                origin[0] + math.cos(angle) * dist,
                origin[1] + math.sin(angle) * dist,
            )
            s = Survivor(*pos)
            if not spritecollideany_walls(s, wall_group):
                survivor_group.add(s)
                all_sprites.add(s, layer=1)
                placed = True
                break
        if not placed:
            s = Survivor(*origin)
            survivor_group.add(s)
            all_sprites.add(s, layer=1)

    game_data.state.survivors_onboard = 0
    apply_passenger_speed_penalty(game_data)


def handle_survivor_zombie_collisions(game_data: GameData, config: dict[str, Any]) -> None:
    if not game_data.stage.rescue_stage:
        return
    survivor_group = game_data.groups.survivor_group
    if not survivor_group:
        return
    zombie_group = game_data.groups.zombie_group
    zombies = [z for z in zombie_group if z.alive()]
    if not zombies:
        return
    zombies.sort(key=lambda s: s.rect.centerx)
    zombie_xs = [z.rect.centerx for z in zombies]
    camera = game_data.camera
    walkable_cells = game_data.layout.walkable_cells
    cell_size = game_data.cell_size

    for survivor in list(survivor_group):
        if not survivor.alive():
            continue
        survivor_radius = survivor.radius
        search_radius = survivor_radius + ZOMBIE_RADIUS
        search_radius_sq = search_radius * search_radius

        min_x = survivor.rect.centerx - search_radius
        max_x = survivor.rect.centerx + search_radius
        start_idx = bisect_left(zombie_xs, min_x)
        collided_zombie: Zombie | None = None
        for idx in range(start_idx, len(zombies)):
            zombie_x = zombie_xs[idx]
            if zombie_x > max_x:
                break
            zombie = zombies[idx]
            if not zombie.alive():
                continue
            dy = zombie.rect.centery - survivor.rect.centery
            if abs(dy) > search_radius:
                continue
            dx = zombie_x - survivor.rect.centerx
            if dx * dx + dy * dy <= search_radius_sq:
                collided_zombie = zombie
                break

        if collided_zombie is None:
            continue
        if not rect_visible_on_screen(camera, survivor.rect):
            spawn_pos = find_nearby_offscreen_spawn_position(
                walkable_cells,
                cell_size,
                camera=camera,
            )
            survivor.teleport(spawn_pos)
            continue
        survivor.kill()
        line = random_survivor_conversion_line(game_data.stage.id)
        if line:
            add_survivor_message(game_data, line)
        new_zombie = _create_zombie(
            config,
            start_pos=survivor.rect.center,
            stage=game_data.stage,
            tracker=collided_zombie.tracker,
            wall_hugging=collided_zombie.wall_hugging,
        )
        zombie_group.add(new_zombie)
        game_data.groups.all_sprites.add(new_zombie, layer=1)
        insert_idx = bisect_left(zombie_xs, new_zombie.rect.centerx)
        zombie_xs.insert(insert_idx, new_zombie.rect.centerx)
        zombies.insert(insert_idx, new_zombie)


def respawn_buddies_near_player(game_data: GameData) -> None:
    """Bring back onboard buddies near the player after losing the car."""
    if game_data.stage.buddy_required_count <= 0:
        return
    count = game_data.state.buddy_onboard
    if count <= 0:
        return

    player = game_data.player
    assert player is not None
    wall_group = game_data.groups.wall_group
    camera = game_data.camera
    walkable_cells = game_data.layout.walkable_cells
    cell_size = game_data.cell_size
    offsets = [
        (BUDDY_RADIUS * 3, 0),
        (-BUDDY_RADIUS * 3, 0),
        (0, BUDDY_RADIUS * 3),
        (0, -BUDDY_RADIUS * 3),
        (0, 0),
    ]
    for _ in range(count):
        if walkable_cells:
            spawn_pos = find_nearby_offscreen_spawn_position(
                walkable_cells,
                cell_size,
                camera=camera,
            )
        else:
            spawn_pos = (int(player.x), int(player.y))
        for dx, dy in offsets:
            candidate = Survivor(player.x + dx, player.y + dy, is_buddy=True)
            if not spritecollideany_walls(candidate, wall_group):
                spawn_pos = (candidate.x, candidate.y)
                break

        buddy = Survivor(*spawn_pos, is_buddy=True)
        buddy.following = True
        game_data.groups.all_sprites.add(buddy, layer=2)
        game_data.groups.survivor_group.add(buddy)
    game_data.state.buddy_onboard = 0



"""Entity-related constants and helpers."""

from __future__ import annotations

from .screen_constants import FPS

# --- Player and buddy settings ---
HUMANOID_RADIUS = 6
PLAYER_RADIUS = HUMANOID_RADIUS
PLAYER_SPEED = 1.4
FOV_RADIUS = 124  # approximate legacy FOV (80) * 1.55 cap
BUDDY_RADIUS = HUMANOID_RADIUS
BUDDY_FOLLOW_SPEED = PLAYER_SPEED * 0.7
HUMANOID_WALL_BUMP_FRAMES = 7

# --- Jump settings ---
PLAYER_JUMP_RANGE = 15  # px (enough to clear one 50px tile)
SURVIVOR_JUMP_RANGE = int(PLAYER_JUMP_RANGE * 0.7)  # px
JUMP_DURATION_MS = 200
JUMP_SCALE_MAX = 0.15
JUMP_SHADOW_OFFSET = 5

# --- Survivor settings (Stage 4) ---
SURVIVOR_RADIUS = HUMANOID_RADIUS
SURVIVOR_APPROACH_RADIUS = 48
SURVIVOR_APPROACH_SPEED = PLAYER_SPEED * 0.35
SURVIVOR_MAX_SAFE_PASSENGERS = 5
SURVIVOR_MIN_SPEED_FACTOR = 0.35

# --- Flashlight settings ---
FLASHLIGHT_WIDTH = 12
FLASHLIGHT_HEIGHT = 10

# --- Shoes settings ---
SHOES_WIDTH = 14
SHOES_HEIGHT = 12

# --- Zombie settings ---
ZOMBIE_RADIUS = HUMANOID_RADIUS
ZOMBIE_SPEED = PLAYER_SPEED * 0.45
ZOMBIE_WANDER_INTERVAL_MS = 5000
ZOMBIE_SIGHT_RANGE = FOV_RADIUS * 1.2
ZOMBIE_TRACKER_SIGHT_RANGE = ZOMBIE_SIGHT_RANGE * 0.2
FAST_ZOMBIE_BASE_SPEED = PLAYER_SPEED * 0.77
ZOMBIE_SEPARATION_DISTANCE = ZOMBIE_RADIUS * 2.2
ZOMBIE_AGING_DURATION_FRAMES = FPS * 60 * 6  # ~6 minutes at target framerate
ZOMBIE_AGING_MIN_SPEED_RATIO = 0.3
ZOMBIE_TRACKER_SCENT_RADIUS = 70
ZOMBIE_TRACKER_FAR_SCENT_RADIUS = 140
ZOMBIE_TRACKER_NEWER_FOOTPRINT_MS = 5000
ZOMBIE_TRACKER_SCENT_TOP_K = 3
ZOMBIE_TRACKER_SCAN_INTERVAL_MS = int(1000 * 30 / FPS)
ZOMBIE_TRACKER_WANDER_INTERVAL_MS = 2500
ZOMBIE_TRACKER_CROWD_BAND_WIDTH = 50
ZOMBIE_TRACKER_CROWD_BAND_LENGTH = 50
ZOMBIE_TRACKER_CROWD_COUNT = 5
ZOMBIE_TRACKER_RELOCK_DELAY_MS = 3000
ZOMBIE_WALL_HUG_SENSOR_DISTANCE = 24
ZOMBIE_WALL_HUG_PROBE_ANGLE_DEG = 45
ZOMBIE_WALL_HUG_PROBE_STEP = 2.0
ZOMBIE_WALL_HUG_TARGET_GAP = 4.0
ZOMBIE_WALL_HUG_LOST_WALL_MS = 2500

# --- Car and fuel settings ---
CAR_WIDTH = 16
CAR_HEIGHT = 24
CAR_SPEED = 2
CAR_HEALTH = 20
CAR_WALL_DAMAGE = 1
FUEL_CAN_WIDTH = 12
FUEL_CAN_HEIGHT = 15

# --- Wall and beam settings ---
INTERNAL_WALL_HEALTH = 40 * 100
INTERNAL_WALL_BEVEL_DEPTH = 6
STEEL_BEAM_HEALTH = int(INTERNAL_WALL_HEALTH * 1.5)
PLAYER_WALL_DAMAGE = 100
BUDDY_WALL_DAMAGE = int(PLAYER_WALL_DAMAGE * 0.7)
ZOMBIE_WALL_DAMAGE = 1

__all__ = [
    "HUMANOID_RADIUS",
    "PLAYER_RADIUS",
    "PLAYER_SPEED",
    "FOV_RADIUS",
    "BUDDY_RADIUS",
    "BUDDY_FOLLOW_SPEED",
    "HUMANOID_WALL_BUMP_FRAMES",
    "SURVIVOR_RADIUS",
    "SURVIVOR_APPROACH_RADIUS",
    "SURVIVOR_APPROACH_SPEED",
    "SURVIVOR_MAX_SAFE_PASSENGERS",
    "SURVIVOR_MIN_SPEED_FACTOR",
    "FLASHLIGHT_WIDTH",
    "FLASHLIGHT_HEIGHT",
    "SHOES_WIDTH",
    "SHOES_HEIGHT",
    "ZOMBIE_RADIUS",
    "ZOMBIE_SPEED",
    "ZOMBIE_WANDER_INTERVAL_MS",
    "ZOMBIE_SIGHT_RANGE",
    "ZOMBIE_TRACKER_SIGHT_RANGE",
    "FAST_ZOMBIE_BASE_SPEED",
    "ZOMBIE_SEPARATION_DISTANCE",
    "ZOMBIE_AGING_DURATION_FRAMES",
    "ZOMBIE_AGING_MIN_SPEED_RATIO",
    "ZOMBIE_TRACKER_SCENT_RADIUS",
    "ZOMBIE_TRACKER_FAR_SCENT_RADIUS",
    "ZOMBIE_TRACKER_NEWER_FOOTPRINT_MS",
    "ZOMBIE_TRACKER_SCENT_TOP_K",
    "ZOMBIE_TRACKER_SCAN_INTERVAL_MS",
    "ZOMBIE_TRACKER_WANDER_INTERVAL_MS",
    "ZOMBIE_TRACKER_CROWD_BAND_WIDTH",
    "ZOMBIE_TRACKER_CROWD_BAND_LENGTH",
    "ZOMBIE_TRACKER_CROWD_COUNT",
    "ZOMBIE_TRACKER_RELOCK_DELAY_MS",
    "ZOMBIE_WALL_HUG_SENSOR_DISTANCE",
    "ZOMBIE_WALL_HUG_PROBE_ANGLE_DEG",
    "ZOMBIE_WALL_HUG_PROBE_STEP",
    "ZOMBIE_WALL_HUG_TARGET_GAP",
    "ZOMBIE_WALL_HUG_LOST_WALL_MS",
    "CAR_WIDTH",
    "CAR_HEIGHT",
    "CAR_SPEED",
    "CAR_HEALTH",
    "CAR_WALL_DAMAGE",
    "FUEL_CAN_WIDTH",
    "FUEL_CAN_HEIGHT",
    "INTERNAL_WALL_HEALTH",
    "INTERNAL_WALL_BEVEL_DEPTH",
    "STEEL_BEAM_HEALTH",
    "PLAYER_WALL_DAMAGE",
    "BUDDY_WALL_DAMAGE",
    "ZOMBIE_WALL_DAMAGE",
]



from __future__ import annotations

from typing import Any

import pygame

from ..colors import DAWN_AMBIENT_PALETTE_KEY, ambient_palette_key_for_flashlights
from ..entities_constants import SURVIVOR_MAX_SAFE_PASSENGERS
from ..models import GameData, Groups, LevelLayout, ProgressState, Stage
from ..entities import Camera
from .ambient import _set_ambient_palette


def initialize_game_state(config: dict[str, Any], stage: Stage) -> GameData:
    """Initialize and return the base game state objects."""
    starts_with_fuel = not stage.requires_fuel
    if stage.endurance_stage:
        starts_with_fuel = False
    starts_with_flashlight = False
    initial_flashlights = 1 if starts_with_flashlight else 0
    initial_palette_key = ambient_palette_key_for_flashlights(initial_flashlights)
    game_state = ProgressState(
        game_over=False,
        game_won=False,
        game_over_message=None,
        game_over_at=None,
        scaled_overview=None,
        overview_created=False,
        footprints=[],
        last_footprint_pos=None,
        footprint_visible_toggle=True,
        elapsed_play_ms=0,
        has_fuel=starts_with_fuel,
        flashlight_count=initial_flashlights,
        shoes_count=0,
        ambient_palette_key=initial_palette_key,
        hint_expires_at=0,
        hint_target_type=None,
        fuel_message_until=0,
        buddy_rescued=0,
        buddy_onboard=0,
        survivors_onboard=0,
        survivors_rescued=0,
        survivor_messages=[],
        survivor_capacity=SURVIVOR_MAX_SAFE_PASSENGERS,
        seed=None,
        endurance_elapsed_ms=0,
        endurance_goal_ms=max(0, stage.endurance_goal_ms),
        dawn_ready=False,
        dawn_prompt_at=None,
        time_accel_active=False,
        last_zombie_spawn_time=0,
        dawn_carbonized=False,
        debug_mode=False,
        show_fps=False,
        falling_zombies=[],
        falling_spawn_carry=0,
        dust_rings=[],
        player_wall_target_cell=None,
        player_wall_target_ttl=0,
    )

    # Create sprite groups
    all_sprites = pygame.sprite.LayeredUpdates()
    wall_group = pygame.sprite.Group()
    zombie_group = pygame.sprite.Group()
    survivor_group = pygame.sprite.Group()

    # Create camera
    cell_size = stage.tile_size
    level_width = stage.grid_cols * cell_size
    level_height = stage.grid_rows * cell_size
    camera = Camera(level_width, level_height)

    # Define level layout (will be filled by blueprint generation)
    field_rect = pygame.Rect(0, 0, level_width, level_height)

    return GameData(
        state=game_state,
        groups=Groups(
            all_sprites=all_sprites,
            wall_group=wall_group,
            zombie_group=zombie_group,
            survivor_group=survivor_group,
        ),
        camera=camera,
        layout=LevelLayout(
            field_rect=field_rect,
            outside_cells=set(),
            walkable_cells=[],
            outer_wall_cells=set(),
            wall_cells=set(),
            pitfall_cells=set(),
            car_walkable_cells=set(),
            fall_spawn_cells=set(),
            bevel_corners={},
        ),
        fog={
            "hatch_patterns": {},
            "overlays": {},
        },
        stage=stage,
        cell_size=cell_size,
        level_width=level_width,
        level_height=level_height,
        fuel=None,
        flashlights=[],
        shoes=[],
    )


def carbonize_outdoor_zombies(game_data: GameData) -> None:
    """Petrify zombies that have already broken through to the exterior."""
    outside_cells = game_data.layout.outside_cells
    if not outside_cells:
        return
    cell_size = game_data.cell_size
    if cell_size <= 0:
        return
    group = game_data.groups.zombie_group
    if not group:
        return
    for zombie in list(group):
        if not zombie.alive():
            continue
        cell = (
            int(zombie.rect.centerx // cell_size),
            int(zombie.rect.centery // cell_size),
        )
        if cell in outside_cells:
            zombie.carbonize()


def update_endurance_timer(game_data: GameData, dt_ms: int) -> None:
    """Advance the endurance countdown and trigger dawn handoff."""
    stage = game_data.stage
    state = game_data.state
    if not stage.endurance_stage:
        return
    if state.endurance_goal_ms <= 0 or dt_ms <= 0:
        return
    state.endurance_elapsed_ms = min(
        state.endurance_goal_ms,
        state.endurance_elapsed_ms + dt_ms,
    )
    if not state.dawn_ready and state.endurance_elapsed_ms >= state.endurance_goal_ms:
        state.dawn_ready = True
        state.dawn_prompt_at = pygame.time.get_ticks()
        _set_ambient_palette(game_data, DAWN_AMBIENT_PALETTE_KEY, force=True)
    if state.dawn_ready:
        carbonize_outdoor_zombies(game_data)
        state.dawn_carbonized = True



from __future__ import annotations

import math
from enum import Enum
from typing import Any

import pygame
import pygame.surfarray as pg_surfarray  # type: ignore
from pygame import sprite, surface

from ..colors import (
    FOOTPRINT_COLOR,
    LIGHT_GRAY,
    ORANGE,
    WHITE,
    YELLOW,
    get_environment_palette,
)
from ..entities import (
    Camera,
    Player,
)
from ..entities_constants import INTERNAL_WALL_BEVEL_DEPTH, ZOMBIE_RADIUS
from ..font_utils import load_font
from ..gameplay_constants import DEFAULT_FLASHLIGHT_SPAWN_COUNT
from ..localization import get_font_settings
from ..localization import translate as tr
from ..models import DustRing, FallingZombie, Footprint, GameData, Stage
from ..render_assets import RenderAssets
from ..render_constants import (
    ENTITY_SHADOW_ALPHA,
    FALLING_DUST_COLOR,
    FALLING_WHIRLWIND_COLOR,
    FALLING_ZOMBIE_COLOR,
    PITFALL_ABYSS_COLOR,
    PITFALL_EDGE_DEPTH_OFFSET,
    PITFALL_EDGE_METAL_COLOR,
    PITFALL_EDGE_STRIPE_COLOR,
    PITFALL_EDGE_STRIPE_SPACING,
    PLAYER_SHADOW_ALPHA_MULT,
    PLAYER_SHADOW_RADIUS_MULT,
)
from .hud import (
    _build_objective_lines,
    _draw_endurance_timer,
    _draw_hint_indicator,
    _draw_inventory_icons,
    _draw_objective,
    _draw_status_bar,
    _draw_survivor_messages,
    _draw_time_accel_indicator,
    _get_fog_scale,
)
from .shadows import (
    _draw_entity_shadows,
    _draw_single_entity_shadow,
    _draw_wall_shadows,
    _get_shadow_layer,
)


def show_message(
    screen: surface.Surface,
    text: str,
    size: int,
    color: tuple[int, int, int],
    position: tuple[int, int],
) -> None:
    try:
        font_settings = get_font_settings()
        font = load_font(font_settings.resource, font_settings.scaled_size(size))
        text_surface = font.render(text, False, color)
        text_rect = text_surface.get_rect(center=position)

        # Add a semi-transparent background rectangle for better visibility
        bg_padding = 15
        bg_rect = text_rect.inflate(bg_padding * 2, bg_padding * 2)
        bg_surface = pygame.Surface((bg_rect.width, bg_rect.height), pygame.SRCALPHA)
        bg_surface.fill((0, 0, 0, 180))  # Black with 180 alpha (out of 255)
        screen.blit(bg_surface, bg_rect.topleft)

        screen.blit(text_surface, text_rect)
    except pygame.error as e:
        print(f"Error rendering font or surface: {e}")


def wrap_long_segment(segment: str, font: pygame.font.Font, max_width: int) -> list[str]:
    lines: list[str] = []
    current = ""
    for char in segment:
        candidate = current + char
        if font.size(candidate)[0] <= max_width or not current:
            current = candidate
        else:
            lines.append(current)
            current = char
    if current:
        lines.append(current)
    return lines


def wrap_text(text: str, font: pygame.font.Font, max_width: int) -> list[str]:
    if max_width <= 0:
        return [text]
    paragraphs = text.splitlines() or [text]
    lines: list[str] = []
    for paragraph in paragraphs:
        if not paragraph:
            lines.append("")
            continue
        words = paragraph.split(" ")
        if len(words) == 1:
            lines.extend(wrap_long_segment(paragraph, font, max_width))
            continue
        current = ""
        for word in words:
            candidate = f"{current} {word}".strip() if current else word
            if font.size(candidate)[0] <= max_width:
                current = candidate
                continue
            if current:
                lines.append(current)
            if font.size(word)[0] <= max_width:
                current = word
            else:
                lines.extend(wrap_long_segment(word, font, max_width))
                current = ""
        if current:
            lines.append(current)
    return lines


def blit_wrapped_text(
    target: surface.Surface,
    text: str,
    font: pygame.font.Font,
    color: tuple[int, int, int],
    topleft: tuple[int, int],
    max_width: int,
) -> None:
    """Render text with simple wrapping constrained to max_width."""

    x, y = topleft
    line_height = font.get_linesize()
    for line in wrap_text(text, font, max_width):
        if not line:
            y += line_height
            continue
        rendered = font.render(line, False, color)
        target.blit(rendered, (x, y))
        y += line_height


def show_message_wrapped(
    screen: surface.Surface,
    text: str,
    size: int,
    color: tuple[int, int, int],
    position: tuple[int, int],
    *,
    max_width: int,
    line_spacing: int = 2,
) -> None:
    try:
        font_settings = get_font_settings()
        font = load_font(font_settings.resource, font_settings.scaled_size(size))
        lines = wrap_text(text, font, max_width)
        if not lines:
            return
        rendered = [font.render(line, False, color) for line in lines]
        max_line_width = max(surface.get_width() for surface in rendered)
        line_height = font.get_linesize()
        total_height = line_height * len(rendered) + line_spacing * (len(rendered) - 1)

        center_x, center_y = position
        top = center_y - total_height // 2

        bg_padding = 15
        bg_width = max_line_width + bg_padding * 2
        bg_height = total_height + bg_padding * 2
        bg_rect = pygame.Rect(0, 0, bg_width, bg_height)
        bg_rect.center = (center_x, center_y)
        bg_surface = pygame.Surface((bg_rect.width, bg_rect.height), pygame.SRCALPHA)
        bg_surface.fill((0, 0, 0, 180))
        screen.blit(bg_surface, bg_rect.topleft)

        y = top
        for text_surface in rendered:
            text_rect = text_surface.get_rect(centerx=center_x, y=y)
            screen.blit(text_surface, text_rect)
            y += line_height + line_spacing
    except pygame.error as e:
        print(f"Error rendering font or surface: {e}")


def draw_pause_overlay(screen: pygame.Surface) -> None:
    screen_width, screen_height = screen.get_size()
    overlay = pygame.Surface((screen_width, screen_height), pygame.SRCALPHA)
    overlay.fill((0, 0, 0, 150))
    pause_radius = 53
    cx = screen_width // 2
    cy = screen_height // 2 - 18
    pygame.draw.circle(
        overlay,
        LIGHT_GRAY,
        (cx, cy),
        pause_radius,
        width=3,
    )
    bar_width = 10
    bar_height = 38
    gap = 12
    pygame.draw.rect(
        overlay,
        LIGHT_GRAY,
        (cx - gap - bar_width, cy - bar_height // 2, bar_width, bar_height),
    )
    pygame.draw.rect(
        overlay,
        LIGHT_GRAY,
        (cx + gap, cy - bar_height // 2, bar_width, bar_height),
    )
    screen.blit(overlay, (0, 0))
    show_message(
        screen,
        tr("hud.paused"),
        18,
        WHITE,
        (screen_width // 2, 28),
    )
    show_message(
        screen,
        tr("hud.pause_hint"),
        16,
        LIGHT_GRAY,
        (screen_width // 2, screen_height // 2 + 70),
    )




def _max_flashlight_pickups() -> int:
    """Return the maximum flashlight pickups available per stage."""
    return max(1, DEFAULT_FLASHLIGHT_SPAWN_COUNT)


class FogProfile(Enum):
    DARK0 = (0, (0, 0, 0, 255))
    DARK1 = (1, (0, 0, 0, 255))
    DARK2 = (2, (0, 0, 0, 255))
    DAWN = (2, (50, 50, 50, 230))

    def __init__(self, flashlight_count: int, color: tuple[int, int, int, int]) -> None:
        self.flashlight_count = flashlight_count
        self.color = color

    def _scale(self, assets: RenderAssets, stage: Stage | None) -> float:
        count = max(0, min(self.flashlight_count, _max_flashlight_pickups()))
        return _get_fog_scale(assets, count)

    @staticmethod
    def _from_flashlight_count(count: int) -> "FogProfile":
        safe_count = max(0, count)
        if safe_count >= 2:
            return FogProfile.DARK2
        if safe_count == 1:
            return FogProfile.DARK1
        return FogProfile.DARK0


def prewarm_fog_overlays(
    fog_data: dict[str, Any],
    assets: RenderAssets,
    *,
    stage: Stage | None = None,
) -> None:
    """Populate fog overlay cache for each reachable flashlight count."""

    for profile in FogProfile:
        _get_fog_overlay_surfaces(
            fog_data,
            assets,
            profile,
            stage=stage,
        )


def _get_hatch_pattern(
    fog_data: dict[str, Any],
    thickness: int,
    *,
    color: tuple[int, int, int, int] | None = None,
) -> surface.Surface:
    """Return cached dot hatch tile surface (Bayer-ordered, optionally chunky)."""
    cache = fog_data.setdefault("hatch_patterns", {})
    key = (thickness, color)
    if key in cache:
        return cache[key]

    spacing = 4
    oversample = 3
    density = max(1, min(thickness, 16))
    pattern_size = spacing * 8
    hi_spacing = spacing * oversample
    hi_pattern_size = pattern_size * oversample
    pattern = pygame.Surface((hi_pattern_size, hi_pattern_size), pygame.SRCALPHA)

    # 8x8 Bayer matrix values 0..63 for ordered dithering
    bayer = [
        [0, 32, 8, 40, 2, 34, 10, 42],
        [48, 16, 56, 24, 50, 18, 58, 26],
        [12, 44, 4, 36, 14, 46, 6, 38],
        [60, 28, 52, 20, 62, 30, 54, 22],
        [3, 35, 11, 43, 1, 33, 9, 41],
        [51, 19, 59, 27, 49, 17, 57, 25],
        [15, 47, 7, 39, 13, 45, 5, 37],
        [63, 31, 55, 23, 61, 29, 53, 21],
    ]
    threshold = int((density / 16) * 64)
    dot_radius = max(
        1,
        min(hi_spacing, int(math.ceil((density / 16) * hi_spacing))),
    )
    dot_color = color or (0, 0, 0, 255)
    for grid_y in range(8):
        for grid_x in range(8):
            if bayer[grid_y][grid_x] < threshold:
                cx = grid_x * hi_spacing + hi_spacing // 2
                cy = grid_y * hi_spacing + hi_spacing // 2
                pygame.draw.circle(pattern, dot_color, (cx, cy), dot_radius)

    if oversample > 1:
        pattern = pygame.transform.smoothscale(pattern, (pattern_size, pattern_size))

    cache[key] = pattern
    return pattern


def _get_fog_overlay_surfaces(
    fog_data: dict[str, Any],
    assets: RenderAssets,
    profile: FogProfile,
    *,
    stage: Stage | None = None,
) -> dict[str, Any]:
    overlays = fog_data.setdefault("overlays", {})
    key = profile
    if key in overlays:
        return overlays[key]

    scale = profile._scale(assets, stage)
    ring_scale = scale
    if profile.flashlight_count >= 2:
        ring_scale += max(0.0, assets.flashlight_hatch_extra_scale)
    max_radius = int(assets.fov_radius * scale)
    padding = 32
    coverage_width = max(assets.screen_width * 2, max_radius * 2)
    coverage_height = max(assets.screen_height * 2, max_radius * 2)
    width = coverage_width + padding * 2
    height = coverage_height + padding * 2
    center = (width // 2, height // 2)

    hard_surface = pygame.Surface((width, height), pygame.SRCALPHA)
    base_color = profile.color
    hard_surface.fill(base_color)
    pygame.draw.circle(hard_surface, (0, 0, 0, 0), center, max_radius)

    ring_surfaces: list[surface.Surface] = []
    for ring in assets.fog_rings:
        ring_surface = pygame.Surface((width, height), pygame.SRCALPHA)
        pattern = _get_hatch_pattern(
            fog_data,
            ring.thickness,
            color=base_color,
        )
        p_w, p_h = pattern.get_size()
        for y in range(0, height, p_h):
            for x in range(0, width, p_w):
                ring_surface.blit(pattern, (x, y))
        radius = int(assets.fov_radius * ring.radius_factor * ring_scale)
        pygame.draw.circle(ring_surface, (0, 0, 0, 0), center, radius)
        ring_surfaces.append(ring_surface)

    combined_surface = hard_surface.copy()
    for ring_surface in ring_surfaces:
        combined_surface.blit(ring_surface, (0, 0))

    visible_fade_surface = _build_flashlight_fade_surface((width, height), center, max_radius)
    combined_surface.blit(visible_fade_surface, (0, 0))

    overlay_entry = {
        "hard": hard_surface,
        "rings": ring_surfaces,
        "combined": combined_surface,
    }
    overlays[key] = overlay_entry
    return overlay_entry


def _build_flashlight_fade_surface(
    size: tuple[int, int],
    center: tuple[int, int],
    max_radius: int,
    *,
    start_ratio: float = 0.2,
    max_alpha: int = 220,
    outer_extension: int = 30,
) -> surface.Surface:
    """Return a radial gradient so flashlight edges softly darken again."""

    width, height = size
    fade_surface = pygame.Surface((width, height), pygame.SRCALPHA)
    fade_surface.fill((0, 0, 0, 0))

    start_radius = max(0.0, min(max_radius, max_radius * start_ratio))
    end_radius = max(start_radius + 1, max_radius + outer_extension)
    fade_range = max(1.0, end_radius - start_radius)

    alpha_view = None
    if pg_surfarray is not None:
        alpha_view = pg_surfarray.pixels_alpha(fade_surface)
    else:  # pragma: no cover - numpy-less fallback
        fade_surface.lock()

    cx, cy = center
    for y in range(height):
        dy = y - cy
        for x in range(width):
            dx = x - cx
            dist = math.hypot(dx, dy)
            if dist > end_radius:
                dist = end_radius
            if dist <= start_radius:
                alpha = 0
            else:
                progress = min(1.0, (dist - start_radius) / fade_range)
                alpha = int(max_alpha * progress)
            if alpha <= 0:
                continue
            if alpha_view is not None:
                alpha_view[x, y] = alpha
            else:
                fade_surface.set_at((x, y), (0, 0, 0, alpha))

    if alpha_view is not None:
        del alpha_view
    else:  # pragma: no cover
        fade_surface.unlock()

    return fade_surface


def _draw_fall_whirlwind(
    screen: surface.Surface,
    camera: Camera,
    center: tuple[int, int],
    progress: float,
    *,
    scale: float = 1.0,
) -> None:
    base_alpha = FALLING_WHIRLWIND_COLOR[3]
    alpha = int(max(0, min(255, base_alpha * (1.0 - progress))))
    if alpha <= 0:
        return
    color = (
        FALLING_WHIRLWIND_COLOR[0],
        FALLING_WHIRLWIND_COLOR[1],
        FALLING_WHIRLWIND_COLOR[2],
        alpha,
    )
    safe_scale = max(0.4, scale)
    swirl_radius = max(2, int(ZOMBIE_RADIUS * 1.1 * safe_scale))
    offset = max(1, int(ZOMBIE_RADIUS * 0.6 * safe_scale))
    size = swirl_radius * 4
    swirl = pygame.Surface((size, size), pygame.SRCALPHA)
    cx = cy = size // 2
    for idx in range(2):
        angle = progress * math.tau * 0.3 + idx * (math.tau / 2)
        ox = int(math.cos(angle) * offset)
        oy = int(math.sin(angle) * offset)
        pygame.draw.circle(swirl, color, (cx + ox, cy + oy), swirl_radius, width=2)
    world_rect = pygame.Rect(0, 0, 1, 1)
    world_rect.center = center
    screen_center = camera.apply_rect(world_rect).center
    screen.blit(swirl, swirl.get_rect(center=screen_center))


def _draw_falling_fx(
    screen: surface.Surface,
    camera: Camera,
    falling_zombies: list[FallingZombie],
    flashlight_count: int,
    dust_rings: list[DustRing],
) -> None:
    if not falling_zombies and not dust_rings:
        return
    now = pygame.time.get_ticks()
    for fall in falling_zombies:
        pre_fx_ms = max(0, fall.pre_fx_ms)
        fall_duration_ms = max(1, fall.fall_duration_ms)
        fall_start = fall.started_at_ms + pre_fx_ms
        impact_at = fall_start + fall_duration_ms
        if now < fall_start:
            if flashlight_count > 0 and pre_fx_ms > 0:
                fx_progress = max(0.0, min(1.0, (now - fall.started_at_ms) / pre_fx_ms))
                # Make the premonition grow with the impending drop scale.
                pre_scale = 1.0 + (0.9 * fx_progress)
                _draw_fall_whirlwind(
                    screen,
                    camera,
                    fall.start_pos,
                    fx_progress,
                    scale=pre_scale,
                )
            continue
        if now >= impact_at:
            continue
        fall_progress = max(0.0, min(1.0, (now - fall_start) / fall_duration_ms))

        if getattr(fall, "mode", "spawn") == "pitfall":
            scale = 1.0 - fall_progress
            scale = scale * scale
            y_offset = 0.0
        else:
            eased = 1.0 - (1.0 - fall_progress) * (1.0 - fall_progress)
            scale = 2.0 - (1.0 * eased)
            # Add an extra vertical drop from above (1.5x wall depth)
            y_offset = -INTERNAL_WALL_BEVEL_DEPTH * 1.5 * (1.0 - eased)

        radius = ZOMBIE_RADIUS * scale
        cx = fall.target_pos[0]
        cy = fall.target_pos[1] + ZOMBIE_RADIUS - radius + y_offset

        world_rect = pygame.Rect(0, 0, radius * 2, radius * 2)
        world_rect.center = (int(cx), int(cy))
        screen_rect = camera.apply_rect(world_rect)
        pygame.draw.circle(
            screen,
            FALLING_ZOMBIE_COLOR,
            screen_rect.center,
            max(1, int(screen_rect.width / 2)),
        )

    for ring in list(dust_rings):
        elapsed = now - ring.started_at_ms
        if elapsed >= ring.duration_ms:
            dust_rings.remove(ring)
            continue
        progress = max(0.0, min(1.0, elapsed / ring.duration_ms))
        alpha = int(max(0, min(255, FALLING_DUST_COLOR[3] * (1.0 - progress))))
        if alpha <= 0:
            continue
        radius = int(ZOMBIE_RADIUS * (0.7 + progress * 1.9))
        color = (
            FALLING_DUST_COLOR[0],
            FALLING_DUST_COLOR[1],
            FALLING_DUST_COLOR[2],
            alpha,
        )
        world_rect = pygame.Rect(0, 0, 1, 1)
        world_rect.center = ring.pos
        screen_center = camera.apply_rect(world_rect).center
        pygame.draw.circle(screen, color, screen_center, radius, width=2)


def _draw_play_area(
    screen: surface.Surface,
    camera: Camera,
    assets: RenderAssets,
    palette: Any,
    field_rect: pygame.Rect,
    outside_cells: set[tuple[int, int]],
    fall_spawn_cells: set[tuple[int, int]],
    pitfall_cells: set[tuple[int, int]],
) -> tuple[int, int, int, int, set[tuple[int, int]]]:
    grid_snap = assets.internal_wall_grid_snap
    xs, ys, xe, ye = (
        field_rect.left,
        field_rect.top,
        field_rect.right,
        field_rect.bottom,
    )
    xs //= grid_snap
    ys //= grid_snap
    xe //= grid_snap
    ye //= grid_snap

    play_area_rect = pygame.Rect(
        xs * grid_snap,
        ys * grid_snap,
        (xe - xs) * grid_snap,
        (ye - ys) * grid_snap,
    )
    play_area_screen_rect = camera.apply_rect(play_area_rect)
    pygame.draw.rect(screen, palette.floor_primary, play_area_screen_rect)

    view_world = pygame.Rect(
        -camera.camera.x,
        -camera.camera.y,
        assets.screen_width,
        assets.screen_height,
    )
    margin = grid_snap * 2
    view_world.inflate_ip(margin * 2, margin * 2)
    min_world_x = max(xs * grid_snap, view_world.left)
    max_world_x = min(xe * grid_snap, view_world.right)
    min_world_y = max(ys * grid_snap, view_world.top)
    max_world_y = min(ye * grid_snap, view_world.bottom)
    start_x = max(xs, int(min_world_x // grid_snap))
    end_x = min(xe, int(math.ceil(max_world_x / grid_snap)))
    start_y = max(ys, int(min_world_y // grid_snap))
    end_y = min(ye, int(math.ceil(max_world_y / grid_snap)))

    for y in range(start_y, end_y):
        for x in range(start_x, end_x):
            if (x, y) in outside_cells:
                lx, ly = (
                    x * grid_snap,
                    y * grid_snap,
                )
                r = pygame.Rect(
                    lx,
                    ly,
                    grid_snap,
                    grid_snap,
                )
                sr = camera.apply_rect(r)
                if sr.colliderect(screen.get_rect()):
                    pygame.draw.rect(screen, palette.outside, sr)
                continue

            if (x, y) in pitfall_cells:
                lx, ly = (
                    x * grid_snap,
                    y * grid_snap,
                )
                r = pygame.Rect(
                    lx,
                    ly,
                    grid_snap,
                    grid_snap,
                )
                sr = camera.apply_rect(r)
                if not sr.colliderect(screen.get_rect()):
                    continue
                pygame.draw.rect(screen, PITFALL_ABYSS_COLOR, sr)

                if (x, y - 1) not in pitfall_cells:
                    edge_h = max(1, INTERNAL_WALL_BEVEL_DEPTH - PITFALL_EDGE_DEPTH_OFFSET)
                    pygame.draw.rect(screen, PITFALL_EDGE_METAL_COLOR, (sr.x, sr.y, sr.w, edge_h))
                    for sx in range(sr.x - edge_h, sr.right, PITFALL_EDGE_STRIPE_SPACING):
                        pygame.draw.line(
                            screen,
                            PITFALL_EDGE_STRIPE_COLOR,
                            (max(sr.x, sx), sr.y),
                            (min(sr.right - 1, sx + edge_h), sr.y + edge_h - 1),
                            width=2,
                        )

                continue

            use_secondary = ((x // 2) + (y // 2)) % 2 == 0
            if (x, y) in fall_spawn_cells:
                color = palette.fall_zone_secondary if use_secondary else palette.fall_zone_primary
            elif not use_secondary:
                continue
            else:
                color = palette.floor_secondary
            lx, ly = (
                x * grid_snap,
                y * grid_snap,
            )
            r = pygame.Rect(
                lx,
                ly,
                grid_snap,
                grid_snap,
            )
            sr = camera.apply_rect(r)
            if sr.colliderect(screen.get_rect()):
                pygame.draw.rect(screen, color, sr)

    return xs, ys, xe, ye, outside_cells


def _draw_footprints(
    screen: surface.Surface,
    camera: Camera,
    assets: RenderAssets,
    footprints: list[Footprint],
    *,
    config: dict[str, Any],
) -> None:
    if not config.get("footprints", {}).get("enabled", True):
        return
    now = pygame.time.get_ticks()
    for fp in footprints:
        if not fp.visible:
            continue
        age = now - fp.time
        fade = 1 - (age / assets.footprint_lifetime_ms)
        fade = max(assets.footprint_min_fade, fade)
        color = tuple(int(c * fade) for c in FOOTPRINT_COLOR)
        fp_rect = pygame.Rect(
            fp.pos[0] - assets.footprint_radius,
            fp.pos[1] - assets.footprint_radius,
            assets.footprint_radius * 2,
            assets.footprint_radius * 2,
        )
        sr = camera.apply_rect(fp_rect)
        if sr.colliderect(screen.get_rect().inflate(30, 30)):
            pygame.draw.circle(screen, color, sr.center, assets.footprint_radius)


def _draw_entities(
    screen: surface.Surface,
    camera: Camera,
    all_sprites: sprite.LayeredUpdates,
    player: Player,
    *,
    has_fuel: bool,
) -> pygame.Rect:
    screen_rect_inflated = screen.get_rect().inflate(100, 100)
    player_screen_rect: pygame.Rect | None = None
    for entity in all_sprites:
        sprite_screen_rect = camera.apply_rect(entity.rect)
        if sprite_screen_rect.colliderect(screen_rect_inflated):
            screen.blit(entity.image, sprite_screen_rect)
        if entity is player:
            player_screen_rect = sprite_screen_rect
            _draw_fuel_indicator(
                screen,
                player_screen_rect,
                has_fuel=has_fuel,
                in_car=player.in_car,
            )
    return player_screen_rect or camera.apply_rect(player.rect)


def _draw_fuel_indicator(
    screen: surface.Surface,
    player_screen_rect: pygame.Rect,
    *,
    has_fuel: bool,
    in_car: bool,
) -> None:
    if not has_fuel or in_car:
        return
    indicator_size = 4
    padding = 1
    indicator_rect = pygame.Rect(
        player_screen_rect.right - indicator_size - padding,
        player_screen_rect.bottom - indicator_size - padding,
        indicator_size,
        indicator_size,
    )
    pygame.draw.rect(screen, YELLOW, indicator_rect)
    pygame.draw.rect(screen, (180, 160, 40), indicator_rect, width=1)


def _draw_fog_of_war(
    screen: surface.Surface,
    camera: Camera,
    assets: RenderAssets,
    fog_surfaces: dict[str, Any],
    fov_target: pygame.sprite.Sprite | None,
    *,
    stage: Stage | None,
    flashlight_count: int,
    dawn_ready: bool,
) -> None:
    if fov_target is None:
        return
    fov_center_on_screen = list(camera.apply(fov_target).center)
    cam_rect = camera.camera
    horizontal_span = camera.width - assets.screen_width
    vertical_span = camera.height - assets.screen_height
    if horizontal_span <= 0 or (cam_rect.x != 0 and cam_rect.x != -horizontal_span):
        fov_center_on_screen[0] = assets.screen_width // 2
    if vertical_span <= 0 or (cam_rect.y != 0 and cam_rect.y != -vertical_span):
        fov_center_on_screen[1] = assets.screen_height // 2
    fov_center_tuple = (int(fov_center_on_screen[0]), int(fov_center_on_screen[1]))
    if dawn_ready:
        profile = FogProfile.DAWN
    else:
        profile = FogProfile._from_flashlight_count(flashlight_count)
    overlay = _get_fog_overlay_surfaces(
        fog_surfaces,
        assets,
        profile,
        stage=stage,
    )
    combined_surface: surface.Surface = overlay["combined"]
    screen.blit(
        combined_surface,
        combined_surface.get_rect(center=fov_center_tuple),
    )


def _draw_need_fuel_message(
    screen: surface.Surface,
    assets: RenderAssets,
    *,
    has_fuel: bool,
    fuel_message_until: int,
    elapsed_play_ms: int,
) -> None:
    if has_fuel or fuel_message_until <= elapsed_play_ms:
        return
    show_message(
        screen,
        tr("hud.need_fuel"),
        18,
        ORANGE,
        (assets.screen_width // 2, assets.screen_height // 2),
    )


def draw(
    assets: RenderAssets,
    screen: surface.Surface,
    game_data: GameData,
    *,
    config: dict[str, Any],
    hint_target: tuple[int, int] | None = None,
    hint_color: tuple[int, int, int] | None = None,
    fps: float | None = None,
) -> None:
    hint_color = hint_color or YELLOW
    state = game_data.state
    player = game_data.player
    if player is None:
        raise ValueError("draw requires an active player on game_data")

    camera = game_data.camera
    stage = game_data.stage
    outside_cells = game_data.layout.outside_cells
    all_sprites = game_data.groups.all_sprites
    has_fuel = state.has_fuel
    flashlight_count = state.flashlight_count
    active_car = game_data.car if game_data.car and game_data.car.alive() else None
    if player.in_car and game_data.car and game_data.car.alive():
        fov_target = game_data.car
    else:
        fov_target = player

    palette = get_environment_palette(state.ambient_palette_key)
    screen.fill(palette.outside)

    _draw_play_area(
        screen,
        camera,
        assets,
        palette,
        game_data.layout.field_rect,
        outside_cells,
        game_data.layout.fall_spawn_cells,
        game_data.layout.pitfall_cells,
    )
    shadow_layer = _get_shadow_layer(screen.get_size())
    shadow_layer.fill((0, 0, 0, 0))
    drew_shadow = _draw_wall_shadows(
        shadow_layer,
        camera,
        wall_cells=game_data.layout.wall_cells,
        wall_group=game_data.groups.wall_group,
        outer_wall_cells=game_data.layout.outer_wall_cells,
        cell_size=game_data.cell_size,
        light_source_pos=(None if (stage and stage.endurance_stage and state.dawn_ready) else fov_target.rect.center)
        if fov_target
        else None,
    )
    drew_shadow |= _draw_entity_shadows(
        shadow_layer,
        camera,
        all_sprites,
        light_source_pos=fov_target.rect.center if fov_target else None,
        exclude_car=active_car if player.in_car else None,
        outside_cells=outside_cells,
        cell_size=game_data.cell_size,
    )
    player_shadow_alpha = max(1, int(ENTITY_SHADOW_ALPHA * PLAYER_SHADOW_ALPHA_MULT))
    player_shadow_radius = int(ZOMBIE_RADIUS * PLAYER_SHADOW_RADIUS_MULT)
    if player.in_car:
        drew_shadow |= _draw_single_entity_shadow(
            shadow_layer,
            camera,
            entity=active_car,
            light_source_pos=fov_target.rect.center if fov_target else None,
            outside_cells=outside_cells,
            cell_size=game_data.cell_size,
            shadow_radius=player_shadow_radius,
            alpha=player_shadow_alpha,
        )
    else:
        drew_shadow |= _draw_single_entity_shadow(
            shadow_layer,
            camera,
            entity=player,
            light_source_pos=fov_target.rect.center if fov_target else None,
            outside_cells=outside_cells,
            cell_size=game_data.cell_size,
            shadow_radius=player_shadow_radius,
            alpha=player_shadow_alpha,
        )
    if drew_shadow:
        screen.blit(shadow_layer, (0, 0))
    _draw_footprints(
        screen,
        camera,
        assets,
        state.footprints,
        config=config,
    )
    _draw_entities(
        screen,
        camera,
        all_sprites,
        player,
        has_fuel=has_fuel,
    )

    _draw_falling_fx(
        screen,
        camera,
        state.falling_zombies,
        state.flashlight_count,
        state.dust_rings,
    )

    _draw_hint_indicator(
        screen,
        camera,
        assets,
        player,
        hint_target,
        hint_color=hint_color,
        stage=stage,
        flashlight_count=flashlight_count,
    )
    _draw_fog_of_war(
        screen,
        camera,
        assets,
        game_data.fog,
        fov_target,
        stage=stage,
        flashlight_count=flashlight_count,
        dawn_ready=state.dawn_ready,
    )
    _draw_need_fuel_message(
        screen,
        assets,
        has_fuel=has_fuel,
        fuel_message_until=state.fuel_message_until,
        elapsed_play_ms=state.elapsed_play_ms,
    )

    objective_lines = _build_objective_lines(
        stage=stage,
        state=state,
        player=player,
        active_car=active_car,
        has_fuel=has_fuel,
        buddy_onboard=state.buddy_onboard,
        buddy_required=stage.buddy_required_count if stage else 0,
        survivors_onboard=state.survivors_onboard,
    )
    if objective_lines:
        _draw_objective(objective_lines, screen=screen)
    _draw_inventory_icons(
        screen,
        assets,
        has_fuel=has_fuel,
        flashlight_count=flashlight_count,
        shoes_count=state.shoes_count,
    )
    _draw_survivor_messages(screen, assets, list(state.survivor_messages))
    _draw_endurance_timer(screen, assets, stage=stage, state=state)
    _draw_time_accel_indicator(screen, assets, stage=stage, state=state)
    _draw_status_bar(
        screen,
        assets,
        config,
        stage=stage,
        seed=state.seed,
        debug_mode=state.debug_mode,
        zombie_group=game_data.groups.zombie_group,
        falling_spawn_carry=state.falling_spawn_carry,
        show_fps=state.show_fps,
        fps=fps,
    )



"""Gameplay-only constants."""

from __future__ import annotations

from ..entities_constants import ZOMBIE_AGING_DURATION_FRAMES

# --- Survivor settings ---
SURVIVOR_SPEED_PENALTY_PER_PASSENGER = 0.08
SURVIVOR_OVERLOAD_DAMAGE_RATIO = 0.2
SURVIVOR_MESSAGE_DURATION_MS = 2000

# --- Footprint settings (gameplay) ---
FOOTPRINT_STEP_DISTANCE = 40
FOOTPRINT_MAX = 320

# --- Zombie settings ---
MAX_ZOMBIES = 400
ZOMBIE_SPAWN_PLAYER_BUFFER = 230
ZOMBIE_TRACKER_AGING_DURATION_FRAMES = ZOMBIE_AGING_DURATION_FRAMES
FALLING_ZOMBIE_PRE_FX_MS = 600
FALLING_ZOMBIE_DURATION_MS = 450
FALLING_ZOMBIE_DUST_DURATION_MS = 220

# --- Car and fuel settings ---
CAR_ZOMBIE_DAMAGE = 1
FUEL_HINT_DURATION_MS = 1600

# --- Wall settings ---
OUTER_WALL_HEALTH = 999999

__all__ = [
    "SURVIVOR_SPEED_PENALTY_PER_PASSENGER",
    "SURVIVOR_OVERLOAD_DAMAGE_RATIO",
    "SURVIVOR_MESSAGE_DURATION_MS",
    "FOOTPRINT_STEP_DISTANCE",
    "FOOTPRINT_MAX",
    "MAX_ZOMBIES",
    "ZOMBIE_SPAWN_PLAYER_BUFFER",
    "ZOMBIE_TRACKER_AGING_DURATION_FRAMES",
    "FALLING_ZOMBIE_PRE_FX_MS",
    "FALLING_ZOMBIE_DURATION_MS",
    "FALLING_ZOMBIE_DUST_DURATION_MS",
    "CAR_ZOMBIE_DAMAGE",
    "FUEL_HINT_DURATION_MS",
    "OUTER_WALL_HEALTH",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pygame
from pygame import surface, time

from ..colors import RED, YELLOW
from ..gameplay_constants import (
    CAR_HINT_DELAY_MS_DEFAULT,
    SURVIVAL_TIME_ACCEL_SUBSTEPS,
    SURVIVAL_TIME_ACCEL_MAX_SUBSTEP,
)
from ..gameplay import (
    MapGenerationError,
    apply_passenger_speed_penalty,
    check_interactions,
    cleanup_survivor_messages,
    generate_level_from_blueprint,
    initialize_game_state,
    maintain_waiting_car_supply,
    nearest_waiting_car,
    place_flashlights,
    place_fuel_can,
    place_shoes,
    process_player_input,
    setup_player_and_cars,
    spawn_initial_zombies,
    spawn_survivors,
    sync_ambient_palette_with_flashlights,
    update_entities,
    update_footprints,
    update_endurance_timer,
)
from ..input_utils import (
    CONTROLLER_BUTTON_DOWN,
    CONTROLLER_DEVICE_ADDED,
    CONTROLLER_DEVICE_REMOVED,
    init_first_controller,
    init_first_joystick,
    is_accel_active,
    is_select_event,
    is_start_event,
    read_gamepad_move,
)
from ..gameplay.spawn import _alive_waiting_cars
from ..world_grid import build_wall_index
from ..localization import translate as tr
from ..models import Stage
from ..render import draw, draw_debug_overview, draw_pause_overlay, prewarm_fog_overlays, show_message, show_message_wrapped
from ..rng import generate_seed, seed_rng
from ..progress import record_stage_clear
from ..screens import (
    ScreenID,
    ScreenTransition,
    nudge_window_scale,
    present,
    sync_window_size,
    toggle_fullscreen,
)

if TYPE_CHECKING:
    from ..render import RenderAssets


def gameplay_screen(
    screen: surface.Surface,
    clock: time.Clock,
    config: dict[str, Any],
    fps: int,
    stage: Stage,
    *,
    show_pause_overlay: bool,
    seed: int | None,
    render_assets: "RenderAssets",
    debug_mode: bool = False,
    show_fps: bool = False,
) -> ScreenTransition:
    """Main gameplay loop that returns the next screen transition."""

    screen_width = screen.get_width()
    screen_height = screen.get_height()
    mouse_hidden = False

    def _set_mouse_hidden(hidden: bool) -> None:
        nonlocal mouse_hidden
        if mouse_hidden == hidden:
            return
        pygame.mouse.set_visible(not hidden)
        mouse_hidden = hidden

    def _finalize(transition: ScreenTransition) -> ScreenTransition:
        _set_mouse_hidden(False)
        return transition

    seed_value = seed if seed is not None else generate_seed()
    applied_seed = seed_rng(seed_value)

    game_data = initialize_game_state(config, stage)
    game_data.state.seed = applied_seed
    game_data.state.debug_mode = debug_mode
    game_data.state.show_fps = show_fps
    if debug_mode and stage.endurance_stage:
        goal_ms = max(0, stage.endurance_goal_ms)
        if goal_ms > 0:
            remaining = 3 * 60 * 1000  # 3 minutes in ms
            game_data.state.endurance_elapsed_ms = max(0, goal_ms - remaining)
            game_data.state.dawn_ready = False
            game_data.state.dawn_prompt_at = None
            game_data.state.dawn_carbonized = False
    prewarm_fog_overlays(
        game_data.fog,
        render_assets,
        stage=stage,
    )
    paused_manual = False
    paused_focus = False
    ignore_focus_loss_until = 0
    debug_overview = False
    controller = init_first_controller()
    joystick = init_first_joystick() if controller is None else None
    _set_mouse_hidden(pygame.mouse.get_focused())

    try:
        layout_data = generate_level_from_blueprint(game_data, config)
    except MapGenerationError:
        # If generation fails after retries, show error and back to title
        screen.fill((0, 0, 0))
        show_message_wrapped(
            screen,
            tr("errors.map_generation_failed"),
            16,
            RED,
            (screen_width // 2, screen_height // 2),
            max_width=screen_width - 40,
        )
        present(screen)
        pygame.time.delay(3000)
        return _finalize(ScreenTransition(ScreenID.TITLE))

    sync_ambient_palette_with_flashlights(game_data, force=True)
    initial_waiting = max(0, stage.waiting_car_target_count)
    player, waiting_cars = setup_player_and_cars(game_data, layout_data, car_count=initial_waiting)
    game_data.player = player
    game_data.waiting_cars = waiting_cars
    game_data.car = None
    # Only top up if initial placement spawned fewer than the intended baseline (shouldn't happen)
    maintain_waiting_car_supply(game_data, minimum=stage.waiting_car_target_count)
    apply_passenger_speed_penalty(game_data)

    spawn_survivors(game_data, layout_data)

    occupied_centers: set[tuple[int, int]] = set()
    cell_size = game_data.cell_size
    if stage.requires_fuel:
        fuel_spawn_count = stage.fuel_spawn_count
        fuel_can = place_fuel_can(
            layout_data["fuel_cells"] or layout_data["walkable_cells"],
            cell_size,
            player,
            cars=game_data.waiting_cars,
            reserved_centers=occupied_centers,
            count=fuel_spawn_count,
        )
        if fuel_can:
            game_data.fuel = fuel_can
            game_data.groups.all_sprites.add(fuel_can, layer=1)
            occupied_centers.add(fuel_can.rect.center)
    flashlight_count = stage.initial_flashlight_count
    flashlights = place_flashlights(
        layout_data["flashlight_cells"] or layout_data["walkable_cells"],
        cell_size,
        player,
        cars=game_data.waiting_cars,
        reserved_centers=occupied_centers,
        count=max(0, flashlight_count),
    )
    game_data.flashlights = flashlights
    game_data.groups.all_sprites.add(flashlights, layer=1)
    for flashlight in flashlights:
        occupied_centers.add(flashlight.rect.center)

    shoes_count = stage.initial_shoes_count
    shoes_list = place_shoes(
        layout_data["shoes_cells"] or layout_data["walkable_cells"],
        cell_size,
        player,
        cars=game_data.waiting_cars,
        reserved_centers=occupied_centers,
        count=max(0, shoes_count),
    )
    game_data.shoes = shoes_list
    game_data.groups.all_sprites.add(shoes_list, layer=1)

    spawn_initial_zombies(game_data, player, layout_data, config)
    update_footprints(game_data, config)
    level_rect = game_data.layout.field_rect
    overview_surface = pygame.Surface((level_rect.width, level_rect.height))
    while True:
        dt = clock.tick(fps) / 1000.0
        current_fps = clock.get_fps()
        if game_data.state.game_over or game_data.state.game_won:
            if game_data.state.game_won:
                record_stage_clear(stage.id)
            if game_data.state.game_over and not game_data.state.game_won:
                if game_data.state.game_over_at is None:
                    game_data.state.game_over_at = pygame.time.get_ticks()
                if pygame.time.get_ticks() - game_data.state.game_over_at < 1000:
                    draw(
                        render_assets,
                        screen,
                        game_data,
                        config=config,
                        hint_color=None,
                        fps=current_fps,
                    )
                    if game_data.state.game_over_message:
                        show_message(
                            screen,
                            game_data.state.game_over_message,
                            18,
                            RED,
                            (screen_width // 2, screen_height // 2 - 24),
                        )
                    present(screen)
                    continue
            return _finalize(
                ScreenTransition(
                    ScreenID.GAME_OVER,
                    stage=stage,
                    game_data=game_data,
                    config=config,
                )
            )

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return _finalize(ScreenTransition(ScreenID.EXIT))
            if event.type in (pygame.WINDOWSIZECHANGED, pygame.VIDEORESIZE):
                sync_window_size(event, game_data=game_data)
                continue
            if event.type == pygame.WINDOWFOCUSLOST:
                now = pygame.time.get_ticks()
                if now >= ignore_focus_loss_until:
                    paused_focus = True
            if event.type == pygame.WINDOWFOCUSGAINED:
                paused_focus = False
            if event.type == pygame.MOUSEBUTTONDOWN:
                paused_focus = False
            if event.type == pygame.JOYDEVICEADDED or (
                CONTROLLER_DEVICE_ADDED is not None and event.type == CONTROLLER_DEVICE_ADDED
            ):
                if controller is None:
                    controller = init_first_controller()
                if controller is None:
                    joystick = init_first_joystick()
            if event.type == pygame.JOYDEVICEREMOVED or (
                CONTROLLER_DEVICE_REMOVED is not None and event.type == CONTROLLER_DEVICE_REMOVED
            ):
                if controller and not controller.get_init():
                    controller = None
                if joystick and not joystick.get_init():
                    joystick = None
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_LEFTBRACKET:
                    nudge_window_scale(0.5, game_data=game_data)
                    continue
                if event.key == pygame.K_RIGHTBRACKET:
                    nudge_window_scale(2.0, game_data=game_data)
                    continue
                if event.key == pygame.K_f:
                    toggle_fullscreen(game_data=game_data)
                    continue
                if event.key == pygame.K_s and (pygame.key.get_mods() & pygame.KMOD_CTRL):
                    state_snapshot = {k: v for k, v in vars(game_data.state).items() if k != "footprints"}
                    print("STATE DEBUG:", state_snapshot)
                    continue
                if debug_mode:
                    if event.key == pygame.K_ESCAPE:
                        return ScreenTransition(ScreenID.TITLE)
                    if event.key == pygame.K_p:
                        paused_manual = not paused_manual
                    if event.key == pygame.K_o:
                        debug_overview = not debug_overview
                    continue
                if paused_manual:
                    if event.key == pygame.K_ESCAPE:
                        return _finalize(ScreenTransition(ScreenID.TITLE))
                    if event.key == pygame.K_p:
                        paused_manual = False
                    continue
                if event.key in (pygame.K_ESCAPE, pygame.K_p):
                    paused_manual = True
                    continue
            if event.type == pygame.JOYBUTTONDOWN or (
                CONTROLLER_BUTTON_DOWN is not None and event.type == CONTROLLER_BUTTON_DOWN
            ):
                if debug_mode:
                    if is_select_event(event):
                        return _finalize(ScreenTransition(ScreenID.TITLE))
                    if is_start_event(event):
                        paused_manual = not paused_manual
                    continue
                if paused_manual:
                    if is_select_event(event):
                        return _finalize(ScreenTransition(ScreenID.TITLE))
                    if is_start_event(event):
                        paused_manual = False
                    continue
                if is_select_event(event) or is_start_event(event):
                    paused_manual = True
                    continue

        _set_mouse_hidden(pygame.mouse.get_focused())

        paused = paused_manual or paused_focus
        if paused:
            draw(
                render_assets,
                screen,
                game_data,
                config=config,
                fps=current_fps,
            )
            if show_pause_overlay:
                draw_pause_overlay(screen)
            present(screen)
            continue

        keys = pygame.key.get_pressed()
        accel_allowed = not (game_data.state.game_over or game_data.state.game_won)
        accel_active = accel_allowed and is_accel_active(keys, controller, joystick)
        game_data.state.time_accel_active = accel_active
        substeps = SURVIVAL_TIME_ACCEL_SUBSTEPS if accel_active else 1
        sub_dt = min(dt, SURVIVAL_TIME_ACCEL_MAX_SUBSTEP) if accel_active else dt
        wall_index = build_wall_index(game_data.groups.wall_group, cell_size=game_data.cell_size)
        for _ in range(substeps):
            player_ref = game_data.player
            if player_ref is None:
                break
            car_ref = game_data.car
            pad_vector = read_gamepad_move(controller, joystick)
            player_dx, player_dy, car_dx, car_dy = process_player_input(
                keys,
                player_ref,
                car_ref,
                shoes_count=game_data.state.shoes_count,
                pad_input=pad_vector,
            )
            update_entities(
                game_data,
                player_dx,
                player_dy,
                car_dx,
                car_dy,
                config,
                wall_index=wall_index,
            )
            update_footprints(game_data, config)
            step_ms = int(sub_dt * 1000)
            if accel_active:
                step_ms = max(1, step_ms)
            game_data.state.elapsed_play_ms += step_ms
            update_endurance_timer(game_data, step_ms)
            cleanup_survivor_messages(game_data.state)
            check_interactions(game_data, config)
            if game_data.state.game_over or game_data.state.game_won:
                break

        player = game_data.player
        if player is None:
            raise ValueError("Player missing from game data")

        if debug_overview:
            draw_debug_overview(
                render_assets,
                screen,
                overview_surface,
                game_data,
                config,
                screen_width=screen_width,
                screen_height=screen_height,
            )
            present(screen)
            continue

        car_hint_conf = config.get("car_hint", {})
        hint_delay = car_hint_conf.get("delay_ms", CAR_HINT_DELAY_MS_DEFAULT)
        elapsed_ms = game_data.state.elapsed_play_ms
        has_fuel = game_data.state.has_fuel
        hint_enabled = car_hint_conf.get("enabled", True) and not stage.endurance_stage
        hint_target = None
        hint_color = YELLOW
        hint_expires_at = game_data.state.hint_expires_at
        hint_target_type = game_data.state.hint_target_type

        active_car = game_data.car if game_data.car and game_data.car.alive() else None
        if hint_enabled:
            if not has_fuel and game_data.fuel and game_data.fuel.alive():
                target_type = "fuel"
            elif not player.in_car and (active_car or _alive_waiting_cars(game_data)):
                target_type = "car"
            else:
                target_type = None

            if target_type != hint_target_type:
                game_data.state.hint_target_type = target_type
                game_data.state.hint_expires_at = elapsed_ms + hint_delay if target_type else 0
                hint_expires_at = game_data.state.hint_expires_at
                hint_target_type = target_type

            if target_type and hint_expires_at and elapsed_ms >= hint_expires_at and not player.in_car:
                if target_type == "fuel" and game_data.fuel and game_data.fuel.alive():
                    hint_target = game_data.fuel.rect.center
                elif target_type == "car":
                    if active_car:
                        hint_target = active_car.rect.center
                    else:
                        waiting_target = nearest_waiting_car(game_data, (player.x, player.y))
                        if waiting_target:
                            hint_target = waiting_target.rect.center

        draw(
            render_assets,
            screen,
            game_data,
            config=config,
            hint_target=hint_target,
            hint_color=hint_color,
            fps=current_fps,
        )
        present(screen)

    # Should not reach here, but return to title if it happens
    return _finalize(ScreenTransition(ScreenID.TITLE))



from __future__ import annotations

import webbrowser
from typing import Any, Sequence

import pygame
from pygame import surface, time

from ..colors import BLACK, GRAY, LIGHT_GRAY, WHITE
from ..font_utils import load_font
from ..localization import get_font_settings, get_language
from ..localization import translate as tr
from ..models import Stage
from ..progress import load_progress
from ..render import blit_wrapped_text, show_message, wrap_text
from ..rng import generate_seed
from ..input_utils import (
    CONTROLLER_BUTTON_DOWN,
    CONTROLLER_BUTTON_DPAD_DOWN,
    CONTROLLER_BUTTON_DPAD_LEFT,
    CONTROLLER_BUTTON_DPAD_RIGHT,
    CONTROLLER_BUTTON_DPAD_UP,
    CONTROLLER_DEVICE_ADDED,
    CONTROLLER_DEVICE_REMOVED,
    init_first_controller,
    init_first_joystick,
    is_confirm_event,
)
from ..screens import (
    ScreenID,
    ScreenTransition,
    nudge_window_scale,
    present,
    sync_window_size,
    toggle_fullscreen,
)

try:  # pragma: no cover - version fallback not critical for tests
    from ..__about__ import __version__
except Exception:  # pragma: no cover - fallback version
    __version__ = "0.0.0-unknown"

MAX_SEED_DIGITS = 19
README_URLS: dict[str, str] = {
    "en": "https://github.com/tos-kamiya/zombie-escape/blob/main/README.md",
    "ja": "https://github.com/tos-kamiya/zombie-escape/blob/main/README-ja_JP.md",
}
STAGE6_URLS: dict[str, str] = {
    "en": "https://github.com/tos-kamiya/zombie-escape/blob/main/docs/stages-6plus.md",
    "ja": "https://github.com/tos-kamiya/zombie-escape/blob/main/docs/stages-6plus-ja_JP.md",
}
UNCLEARED_STAGE_COLOR: tuple[int, int, int] = (220, 80, 80)


def _open_readme_link(*, use_stage6: bool = False) -> None:
    """Open the GitHub README or Stage 6+ guide for the active UI language."""

    language = get_language()
    if use_stage6:
        url = STAGE6_URLS.get(language, STAGE6_URLS["en"])
    else:
        url = README_URLS.get(language, README_URLS["en"])
    try:
        webbrowser.open(url, new=0, autoraise=True)
    except Exception as exc:  # pragma: no cover - best effort only
        print(f"Unable to open README URL {url}: {exc}")


def _generate_auto_seed_text() -> str:
    raw = generate_seed()
    trimmed = raw // 100  # drop lower 2 digits for stability
    return str(trimmed % 100000).zfill(5)


def title_screen(
    screen: surface.Surface,
    clock: time.Clock,
    config: dict[str, Any],
    fps: int,
    *,
    stages: Sequence[Stage],
    default_stage_id: str,
    screen_size: tuple[int, int],
    seed_text: str | None = None,
    seed_is_auto: bool = False,
) -> ScreenTransition:
    """Display the title menu and return the selected transition."""

    width, height = screen_size
    stage_options_all: list[dict] = [
        {"type": "stage", "stage": stage, "available": stage.available} for stage in stages if stage.available
    ]
    page_size = 5
    stage_pages = [stage_options_all[i : i + page_size] for i in range(0, len(stage_options_all), page_size)]
    action_options: list[dict[str, Any]] = [
        {"type": "settings"},
        {"type": "readme"},
        {"type": "quit"},
    ]
    generated = seed_text is None
    current_seed_text = seed_text if seed_text is not None else _generate_auto_seed_text()
    current_seed_auto = seed_is_auto or generated
    stage_progress, _ = load_progress()

    def _page_available(page_index: int) -> bool:
        if page_index <= 0:
            return True
        required = stage_options_all[:page_size]
        return all(stage_progress.get(option["stage"].id, 0) > 0 for option in required)

    current_page = 0
    if stage_options_all:
        for idx, opt in enumerate(stage_options_all):
            if opt["stage"].id == default_stage_id:
                target_page = idx // page_size
                if _page_available(target_page):
                    current_page = target_page
                break

    def _build_options(page_index: int) -> tuple[list[dict], list[dict]]:
        page_index = max(0, min(page_index, len(stage_pages) - 1))
        stage_options = stage_pages[page_index] if stage_pages else []
        options = list(stage_options) + action_options
        return options, stage_options

    options, stage_options = _build_options(current_page)
    selected_stage_index = next(
        (i for i, opt in enumerate(options) if opt["type"] == "stage" and opt["stage"].id == default_stage_id),
        0,
    )
    selected = min(selected_stage_index, len(options) - 1)
    controller = init_first_controller()
    joystick = init_first_joystick() if controller is None else None

    while True:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                return ScreenTransition(
                    ScreenID.EXIT,
                    seed_text=current_seed_text,
                    seed_is_auto=current_seed_auto,
                )
            if event.type in (pygame.WINDOWSIZECHANGED, pygame.VIDEORESIZE):
                sync_window_size(event)
                continue
            if event.type == pygame.JOYDEVICEADDED or (
                CONTROLLER_DEVICE_ADDED is not None and event.type == CONTROLLER_DEVICE_ADDED
            ):
                if controller is None:
                    controller = init_first_controller()
                if controller is None:
                    joystick = init_first_joystick()
            if event.type == pygame.JOYDEVICEREMOVED or (
                CONTROLLER_DEVICE_REMOVED is not None and event.type == CONTROLLER_DEVICE_REMOVED
            ):
                if controller and not controller.get_init():
                    controller = None
                if joystick and not joystick.get_init():
                    joystick = None
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_BACKSPACE:
                    current_seed_text = _generate_auto_seed_text()
                    current_seed_auto = True
                    continue
                if event.unicode and event.unicode.isdigit():
                    if current_seed_auto:
                        current_seed_text = ""
                        current_seed_auto = False
                    if len(current_seed_text) < MAX_SEED_DIGITS:
                        current_seed_text += event.unicode
                    continue
                if event.key == pygame.K_LEFTBRACKET:
                    nudge_window_scale(0.5)
                    continue
                if event.key == pygame.K_RIGHTBRACKET:
                    nudge_window_scale(2.0)
                    continue
                if event.key == pygame.K_f:
                    toggle_fullscreen()
                    continue
                if event.key == pygame.K_LEFT:
                    if current_page > 0:
                        current_page -= 1
                        options, stage_options = _build_options(current_page)
                        selected = 0
                    continue
                if event.key == pygame.K_RIGHT:
                    if current_page < len(stage_pages) - 1 and _page_available(current_page + 1):
                        current_page += 1
                        options, stage_options = _build_options(current_page)
                        selected = 0
                    continue
                if event.key in (pygame.K_UP, pygame.K_w):
                    selected = (selected - 1) % len(options)
                elif event.key in (pygame.K_DOWN, pygame.K_s):
                    selected = (selected + 1) % len(options)
                elif event.key in (pygame.K_RETURN, pygame.K_SPACE):
                    current = options[selected]
                    if current["type"] == "stage" and current.get("available"):
                        seed_value = int(current_seed_text) if current_seed_text else None
                        return ScreenTransition(
                            ScreenID.GAMEPLAY,
                            stage=current["stage"],
                            seed=seed_value,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                    if current["type"] == "settings":
                        return ScreenTransition(
                            ScreenID.SETTINGS,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                    if current["type"] == "readme":
                        _open_readme_link(use_stage6=current_page > 0)
                        continue
                    if current["type"] == "quit":
                        return ScreenTransition(
                            ScreenID.EXIT,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
            if event.type == pygame.JOYBUTTONDOWN or (
                CONTROLLER_BUTTON_DOWN is not None and event.type == CONTROLLER_BUTTON_DOWN
            ):
                if is_confirm_event(event):
                    current = options[selected]
                    if current["type"] == "stage" and current.get("available"):
                        seed_value = int(current_seed_text) if current_seed_text else None
                        return ScreenTransition(
                            ScreenID.GAMEPLAY,
                            stage=current["stage"],
                            seed=seed_value,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                    if current["type"] == "settings":
                        return ScreenTransition(
                            ScreenID.SETTINGS,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                    if current["type"] == "readme":
                        _open_readme_link(use_stage6=current_page > 0)
                        continue
                    if current["type"] == "quit":
                        return ScreenTransition(
                            ScreenID.EXIT,
                            seed_text=current_seed_text,
                            seed_is_auto=current_seed_auto,
                        )
                if CONTROLLER_BUTTON_DOWN is not None and event.type == CONTROLLER_BUTTON_DOWN:
                    if CONTROLLER_BUTTON_DPAD_UP is not None and event.button == CONTROLLER_BUTTON_DPAD_UP:
                        selected = (selected - 1) % len(options)
                    if CONTROLLER_BUTTON_DPAD_DOWN is not None and event.button == CONTROLLER_BUTTON_DPAD_DOWN:
                        selected = (selected + 1) % len(options)
                    if CONTROLLER_BUTTON_DPAD_LEFT is not None and event.button == CONTROLLER_BUTTON_DPAD_LEFT:
                        if current_page > 0:
                            current_page -= 1
                            options, stage_options = _build_options(current_page)
                            selected = 0
                    if CONTROLLER_BUTTON_DPAD_RIGHT is not None and event.button == CONTROLLER_BUTTON_DPAD_RIGHT:
                        if current_page < len(stage_pages) - 1 and _page_available(current_page + 1):
                            current_page += 1
                            options, stage_options = _build_options(current_page)
                            selected = 0
            if event.type == pygame.JOYHATMOTION:
                hat_x, hat_y = event.value
                if hat_y == 1:
                    selected = (selected - 1) % len(options)
                elif hat_y == -1:
                    selected = (selected + 1) % len(options)
                if hat_x == -1:
                    if current_page > 0:
                        current_page -= 1
                        options, stage_options = _build_options(current_page)
                        selected = 0
                elif hat_x == 1:
                    if current_page < len(stage_pages) - 1 and _page_available(current_page + 1):
                        current_page += 1
                        options, stage_options = _build_options(current_page)
                        selected = 0

        screen.fill(BLACK)
        title_text = tr("game.title")
        show_message(screen, title_text, 32, LIGHT_GRAY, (width // 2, 40))

        try:
            font_settings = get_font_settings()
            title_font = load_font(font_settings.resource, font_settings.scaled_size(32))
            option_font = load_font(font_settings.resource, font_settings.scaled_size(14))
            desc_font = load_font(font_settings.resource, font_settings.scaled_size(11))
            section_font = load_font(font_settings.resource, font_settings.scaled_size(13))
            hint_font = load_font(font_settings.resource, font_settings.scaled_size(11))

            row_height = 20
            list_column_x = 24
            list_column_width = width // 2 - 36
            info_column_x = width // 2 + 12
            info_column_width = width - info_column_x - 24
            section_top = 70
            highlight_color = (70, 70, 70)

            stage_count = len(stage_options)
            # resource_count = len(options) - stage_count

            stage_header_text = tr("menu.sections.stage_select")
            show_page_arrows = len(stage_pages) > 1 and _page_available(1)
            if show_page_arrows:
                left_arrow = "<- " if current_page > 0 else ""
                right_arrow = " ->" if current_page < len(stage_pages) - 1 and _page_available(current_page + 1) else ""
                stage_header_text = f"{left_arrow}{stage_header_text}{right_arrow}"
            stage_header = section_font.render(stage_header_text, False, LIGHT_GRAY)
            stage_header_pos = (list_column_x, section_top)
            screen.blit(stage_header, stage_header_pos)
            stage_rows_start = stage_header_pos[1] + stage_header.get_height() + 6
            action_header = section_font.render(tr("menu.sections.resources"), False, LIGHT_GRAY)
            action_header_pos = (
                list_column_x,
                stage_rows_start + stage_count * row_height + 14,
            )
            screen.blit(action_header, action_header_pos)
            action_rows_start = action_header_pos[1] + action_header.get_height() + 6

            for idx, option in enumerate(stage_options):
                row_top = stage_rows_start + idx * row_height
                highlight_rect = pygame.Rect(list_column_x, row_top - 2, list_column_width, row_height)
                cleared = stage_progress.get(option["stage"].id, 0) > 0
                base_color = WHITE if cleared else UNCLEARED_STAGE_COLOR
                color = base_color
                if idx == selected:
                    pygame.draw.rect(screen, highlight_color, highlight_rect)
                label = option["stage"].name
                if not option.get("available"):
                    locked_suffix = tr("menu.locked_suffix")
                    label = f"{label} {locked_suffix}"
                    color = GRAY
                text_surface = option_font.render(label, False, color)
                text_rect = text_surface.get_rect(
                    topleft=(
                        list_column_x + 8,
                        row_top + (row_height - text_surface.get_height()) // 2,
                    )
                )
                screen.blit(text_surface, text_rect)

            for idx, option in enumerate(action_options):
                option_idx = stage_count + idx
                row_top = action_rows_start + idx * row_height
                highlight_rect = pygame.Rect(list_column_x, row_top - 2, list_column_width, row_height)
                is_selected = option_idx == selected
                if is_selected:
                    pygame.draw.rect(screen, highlight_color, highlight_rect)
                if option["type"] == "settings":
                    label = tr("menu.settings")
                elif option["type"] == "readme":
                    label_key = "menu.readme_stage6" if current_page > 0 else "menu.readme"
                    label = f"> {tr(label_key)}"
                else:
                    label = tr("menu.quit")
                color = WHITE
                text_surface = option_font.render(label, False, color)
                text_rect = text_surface.get_rect(
                    topleft=(
                        list_column_x + 8,
                        row_top + (row_height - text_surface.get_height()) // 2,
                    )
                )
                screen.blit(text_surface, text_rect)

            current = options[selected]
            desc_area_top = section_top
            if current["type"] == "stage":
                desc_color = WHITE if current.get("available") else GRAY
                blit_wrapped_text(
                    screen,
                    current["stage"].description,
                    desc_font,
                    desc_color,
                    (info_column_x, desc_area_top),
                    info_column_width,
                )

            option_help_top = desc_area_top
            help_text = ""
            if current["type"] == "settings":
                help_text = tr("menu.option_help.settings")
            elif current["type"] == "quit":
                help_text = tr("menu.option_help.quit")
            elif current["type"] == "readme":
                help_key = "menu.option_help.readme_stage6" if current_page > 0 else "menu.option_help.readme"
                help_text = tr(help_key)

            if help_text:
                blit_wrapped_text(
                    screen,
                    help_text,
                    desc_font,
                    WHITE,
                    (info_column_x, option_help_top),
                    info_column_width,
                )

            hint_lines = [tr("menu.hints.navigate")]
            if len(stage_pages) > 1 and _page_available(1):
                hint_lines.append(tr("menu.hints.page_switch"))
            hint_lines.append(tr("menu.hints.confirm"))
            hint_line_height = hint_font.get_linesize()
            # hint_block_height = len(hint_lines) * hint_line_height
            hint_start_y = action_header_pos[1]
            for offset, line in enumerate(hint_lines):
                hint_surface = hint_font.render(line, False, WHITE)
                hint_rect = hint_surface.get_rect(topleft=(info_column_x, hint_start_y + offset * hint_line_height))
                screen.blit(hint_surface, hint_rect)

            seed_value_display = current_seed_text if current_seed_text else tr("menu.seed_empty")
            seed_label = tr("menu.seed_label", value=seed_value_display)
            seed_surface = hint_font.render(seed_label, False, LIGHT_GRAY)
            seed_offset_y = hint_line_height
            seed_rect = seed_surface.get_rect(bottomleft=(info_column_x, height - 30 + seed_offset_y))
            screen.blit(seed_surface, seed_rect)

            seed_hint = tr("menu.seed_hint")
            seed_hint_lines = wrap_text(seed_hint, hint_font, info_column_width)
            seed_hint_height = len(seed_hint_lines) * hint_line_height
            seed_hint_top = seed_rect.top - 4 - seed_hint_height
            blit_wrapped_text(
                screen,
                seed_hint,
                hint_font,
                LIGHT_GRAY,
                (info_column_x, seed_hint_top),
                info_column_width,
            )

            title_surface = title_font.render(title_text, False, LIGHT_GRAY)
            title_rect = title_surface.get_rect(center=(width // 2, 40))
            version_font = load_font(font_settings.resource, font_settings.scaled_size(15))
            version_surface = version_font.render(f"v{__version__}", False, LIGHT_GRAY)
            version_rect = version_surface.get_rect(topleft=(title_rect.right + 4, title_rect.bottom - 4))
            screen.blit(version_surface, version_rect)

        except pygame.error as e:
            print(f"Error rendering title screen: {e}")

        present(screen)
        clock.tick(fps)



"""Screen framework utilities for zombie_escape.

This module defines common types that allow each screen (title, settings,
gameplay, game over, etc.) to communicate transitions in a consistent way.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING

import pygame
from pygame import surface

try:  # pragma: no cover - version fallback not critical for tests
    from ..__about__ import __version__
except Exception:  # pragma: no cover - fallback version
    __version__ = "0.0.0-unknown"
from ..screen_constants import (
    DEFAULT_WINDOW_SCALE,
    SCREEN_HEIGHT,
    SCREEN_WIDTH,
    WINDOW_SCALE_MAX,
    WINDOW_SCALE_MIN,
)

if TYPE_CHECKING:  # pragma: no cover - typing only
    from ..models import GameData, Stage


class ScreenID(Enum):
    """Identifiers for the major screens in the game."""

    TITLE = "title"
    SETTINGS = "settings"
    GAMEPLAY = "gameplay"
    GAME_OVER = "game_over"
    EXIT = "exit"


@dataclass(frozen=True)
class ScreenTransition:
    """Represents the next screen to display and optional payload data."""

    next_screen: ScreenID
    stage: Stage | None = None
    game_data: GameData | None = None
    config: dict | None = None
    seed: int | None = None
    seed_text: str | None = None
    seed_is_auto: bool = False


current_window_scale = DEFAULT_WINDOW_SCALE  # Applied to the OS window only
current_maximized = False
last_window_scale = DEFAULT_WINDOW_SCALE
current_window_size = (
    int(SCREEN_WIDTH * DEFAULT_WINDOW_SCALE),
    int(SCREEN_HEIGHT * DEFAULT_WINDOW_SCALE),
)
last_logged_window_size = current_window_size

__all__ = [
    "ScreenID",
    "ScreenTransition",
    "present",
    "apply_window_scale",
    "nudge_window_scale",
    "toggle_fullscreen",
    "sync_window_size",
]


def present(logical_surface: surface.Surface) -> None:
    """Scale the logical surface directly to the window and flip buffers."""
    window = pygame.display.get_surface()
    if window is None:
        return
    window_size = _fetch_window_size(window)
    _update_window_size(window_size, source="frame")
    logical_size = logical_surface.get_size()
    if window_size == logical_size:
        window.blit(logical_surface, (0, 0))
    else:
        # Preserve aspect ratio with letterboxing.
        scale_x = window_size[0] / max(1, logical_size[0])
        scale_y = window_size[1] / max(1, logical_size[1])
        scale = min(scale_x, scale_y)
        scaled_width = max(1, int(logical_size[0] * scale))
        scaled_height = max(1, int(logical_size[1] * scale))
        window.fill((0, 0, 0))
        if (scaled_width, scaled_height) == logical_size:
            scaled_surface = logical_surface
        else:
            scaled_surface = pygame.transform.scale(logical_surface, (scaled_width, scaled_height))
        offset_x = (window_size[0] - scaled_width) // 2
        offset_y = (window_size[1] - scaled_height) // 2
        window.blit(scaled_surface, (offset_x, offset_y))
    pygame.display.flip()


def apply_window_scale(scale: float, *, game_data: "GameData | None" = None) -> surface.Surface:
    """Resize the OS window; logical render surface stays constant."""
    global current_window_scale, current_maximized, last_window_scale

    clamped_scale = max(WINDOW_SCALE_MIN, min(WINDOW_SCALE_MAX, scale))
    current_window_scale = clamped_scale
    last_window_scale = clamped_scale
    current_maximized = False

    window_width = max(1, int(SCREEN_WIDTH * current_window_scale))
    window_height = max(1, int(SCREEN_HEIGHT * current_window_scale))

    new_window = pygame.display.set_mode((window_width, window_height), pygame.RESIZABLE)
    _update_window_size((window_width, window_height), source="apply_scale")
    _update_window_caption()

    if game_data is not None:
        game_data.state.overview_created = False

    return new_window


def nudge_window_scale(multiplier: float, *, game_data: "GameData | None" = None) -> surface.Surface:
    """Scale the window relative to the current zoom level."""
    target_scale = current_window_scale * multiplier
    return apply_window_scale(target_scale, game_data=game_data)


def toggle_fullscreen(*, game_data: "GameData | None" = None) -> surface.Surface | None:
    """Toggle a maximized window without persisting the setting."""
    global current_maximized, last_window_scale
    if current_maximized:
        current_maximized = False
        window_width = max(1, int(SCREEN_WIDTH * last_window_scale))
        window_height = max(1, int(SCREEN_HEIGHT * last_window_scale))
        window = pygame.display.set_mode((window_width, window_height), pygame.RESIZABLE)
        _restore_window()
        _update_window_caption()
        _update_window_size((window_width, window_height), source="toggle_windowed")
    else:
        last_window_scale = current_window_scale
        current_maximized = True
        window = pygame.display.set_mode(_fetch_window_size(None), pygame.RESIZABLE)
        _maximize_window()
        window_width, window_height = _fetch_window_size(window)
        _update_window_caption()
        _update_window_size((window_width, window_height), source="toggle_fullscreen")
    pygame.mouse.set_visible(not current_maximized)
    if game_data is not None:
        game_data.state.overview_created = False
    return window


def sync_window_size(event: pygame.event.Event, *, game_data: "GameData | None" = None) -> None:
    """Synchronize tracked window size with SDL window events."""
    global current_window_scale, last_window_scale
    size = getattr(event, "size", None)
    if not size:
        width = getattr(event, "x", None)
        height = getattr(event, "y", None)
        if width is not None and height is not None:
            size = (width, height)
    if not size:
        return
    window_width, window_height = _normalize_window_size(size)
    _update_window_size((window_width, window_height), source="window_event")
    if not current_maximized:
        scale_x = window_width / max(1, SCREEN_WIDTH)
        scale_y = window_height / max(1, SCREEN_HEIGHT)
        scale = max(WINDOW_SCALE_MIN, min(WINDOW_SCALE_MAX, min(scale_x, scale_y)))
        current_window_scale = scale
        last_window_scale = scale
    _update_window_caption()
    if game_data is not None:
        game_data.state.overview_created = False


def _fetch_window_size(window: surface.Surface | None) -> tuple[int, int]:
    if hasattr(pygame.display, "get_window_size"):
        size = pygame.display.get_window_size()
        if size != (0, 0):
            return _normalize_window_size(size)
    if window is not None:
        return _normalize_window_size(window.get_size())
    window_width = max(1, int(SCREEN_WIDTH * last_window_scale))
    window_height = max(1, int(SCREEN_HEIGHT * last_window_scale))
    return window_width, window_height


def _normalize_window_size(size: tuple[int, int]) -> tuple[int, int]:
    width = max(1, int(size[0]))
    height = max(1, int(size[1]))
    return width, height


def _update_window_size(size: tuple[int, int], *, source: str) -> None:
    global current_window_size, last_logged_window_size
    current_window_size = size
    if size != last_logged_window_size:
        print(f"WINDOW_SIZE {source}={size[0]}x{size[1]}")
        last_logged_window_size = size


def _update_window_caption() -> None:
    pygame.display.set_caption("Zombie Escape")


def _maximize_window() -> None:
    try:
        from pygame import _sdl2 as sdl2  # type: ignore[import-not-found]
    except Exception:
        return
    try:
        window = sdl2.Window.from_display_module()
    except Exception:
        return
    try:
        window.maximize()
    except Exception:
        return


def _restore_window() -> None:
    try:
        from pygame import _sdl2 as sdl2  # type: ignore[import-not-found]
    except Exception:
        return
    try:
        window = sdl2.Window.from_display_module()
    except Exception:
        return
    try:
        window.restore()
    except Exception:
        return
